# Complete and Improved FPGA Implementation of Classic McEliece

We present the first specification-compliant constant-time FPGA implementation 
of the Classic McEliece cryptosystem from the third-round of NIST’s 
Post-Quantum Cryptography standardization process. In particular, 
we present the first complete implementation including encapsulation and 
decapsulation modules as well as key generation with seed expansion. 
All the hardware modules are parametrizable, at compile time, with security level 
and performance parameters. This source code accompanies the paper “Complete and Improved FPGA Implementation of Classic McEliece” [1],
However, for copyright reasons, we cannot provide the source code for the three systemizers described in the paper
(hybrid early-abort systemizer (HEA), single-pass early-abort systemizer (SPEA), and dual-pass earlyabort systemizer (DPEA))
but only a re-implementation of the SPEA variant (SYSTEMIZER\_SPEA\_REDO).
We refer to [1] and [2] for the original and up-to-date version of the paper.


# Project's Structure
The project is a GNU Makefile project.
The modules are written in Verilog and parameterized with built-in Verilog mechanisms, SageMath, and Python scripts.
The makefile at the top includes targets for the top modules mentioned in the paper:
seeded key generation, encapsulation, decapsulation, and the joint design.
The makefile project automates the source and testbench generation, 
the simulation and synthesis (for Xilinx devices), 
as well as the generation of known answer tests (KATs).
The makefile structure enables the automated building process
for the various specified Classic McEliece parameters sets, 
module configarations, and FPGA models. 

The purpose of the individual folders are as follow:

| Folder          | Description                                                                |
| --------------- | -------------------------------------------------------------------------- |
| `build/`        | Build folder including generated sources, KAT files, and results.          |
| `config/`       | Includes parameter definitions, general tool and synthesis configurations. |
| `kat/`          | Known answer test generation and verification scripts.                     |
| `modules/`      | Hardware modules source files.                                             |
| `Makefile`      | Top Makefile.                                                              |


# How to use the Makefile
The top Makefile includes the cumulated targets for the sub-targets defined in the sub-Makefiles in the individual sub-folders. 
All targtes have their dependencies so that you can build arbitrary targets and all dependencies are build recursively. 
This enables a parallel build process as well. 

To generate targets for all parameter sets, the Makefiles use the second expansion feature of make, c.f., [3].
Therewith, we are able to generate targets for all combinations, e.g., 
over all Classic McEliece parameter sets, design simulators, or supported FPGA models, 
for design source generation, simulation execution, or synthesis runs.

## Available Targets
| Target                                                                             | Description                                                                                                     |
| ------                                                                             | -----------                                                                                                     |
| `all`                                                                              | Generate sources and run the simulation and the syntheses for all topmodules. Same as `results`.                |
| `encap`, `decap`, `keygen`, `joint_design`                                         | Build sources, testbenches, and run simulations and syntheses for the topmodules. Same as `<topmodule>-results` |
| `sources`                                                                          | Build the sources for all topmodules.                                                                           |
| `encap-sources`, `decap-sources`, `keygen-sources`, `joint_design-sources`         | Build the sources for a topmodule.                                                                              |
| `testbench`                                                                        | Build the testbench sources for all topmodules.                                                                 |
| `encap-testbench`, `decap-testbench`, `keygen-testbench`, `joint_design-testbench` | Build the testbench sources for a topmodule.                                                                    |
| `sim`                                                                              | Run the simulations for all topmodules.                                                                         |
| `encap-sim`, `decap-sim`, `keygen-sim`, `joint_design-sim`                         | Run the simulations for a topmodule.                                                                            |
| `synthesis`                                                                        | Run the syntheses for all topmodules.                                                                           |
| `encap-synthesis`, `decap-synthesis`, `keygen-synthesis`, `joint_design-synthesis` | Run the syntheses for a topmodule.                                                                              |
| `results`                                                                          | Generate sources and run the simulation and the syntheses for all topmodules. Same as `all`.                    |
| `encap-results`, `decap-results`, `keygen-results`, `joint_design-results`         | Generate sources and run the simulation and the syntheses for a topmodule.                                      |
| `kat`                                                                              | Generate Known Answer Tests (KATs) for Classic McEliece.                                                        |
| `clean`                                                                            | Remove the entire build directory.                                                                              |

## Makefile Variables
| Variable               | Description                                                                                                                                                                                                                                                                                         |
| --------               | -----------                                                                                                                                                                                                                                                                                         |
| `BUILD_DIR=<path>`     | Set the build directory for all targets.<br>Default: `build`                                                                                                                                                                                                                                        |
| `PAR_SETS=<list>`      | List of target Classic McEliece parameter sets.<br>Predefined from specification: "`mceliece348864` `mceliece460896` `mceliece6688128` `mceliece6960119` `mceliece8192128`"<br> Find definitions in `config/parameters.mk`. |
| `XILINX_MODELS=<list>` | List of target Xilinx FPGA models for the analyses.<br>Default: "`artix7_200t` `artix7_100t` `artix7_35t`"<br>Find definitions in `config/Xilinx/models.mk`.                                                                                                                                        |
| `SYSTEMIZER=<name>`    | Name of the desired systemizer.<br>Default: `SYSTEMIZER_SPEA_REDO`.<br>Available systemizers: `SYSTEMIZER_SPGF2` `SYSTEMIZER_SPGF2_REDO` `SYSTEMIZER_SPGFX` `SYSTEMIZER_SPGFXA` `SYSTEMIZER_SPEA_REDO`.                                             |
| `NPROC=<num>`          | Set number of parallel make jobs.                                                                                                                                                                                                                                                                   |

## Variables and Required Versions of Tools
Tools are defined in `config/tools.mk`.

| Variable                          | Description                                  | Min. Version |
| --------                          | ----------                                   |         ---- |
| `MAKE=<[path to] make>`           | Set GNU Make binary. Default: `make`.        |          4.3 |
| `PYTHON=<[path to] python>`       | Set Python binary. Default: `python3`.       |          3.8 |
| `SAGE=<[path to] sage>`           | Set SageMath binary. Default: `sage`.        |          9.3 |
| `VERILATOR=<[path to] verilator>` | Set Verilator binary. Default: `verilator`.  |        4.216 |
| `VIVADO=<[path to] vivado>`       | Set Xilinx Vivado binary. Default: `vivado`. |     2019.2.1 |

## Examples
Get help in the command line:
```
make help
```

Run the simulations and syntheses for all specified Classic McEliece parameter sets:
```
make all
```

Run the simulations and syntheses just for `mceliece348864`:
```
make all PAR_SETS=mceliece348864
```

Run the simulation for the encapsulation module for `mceliece6688128` and `mceliece8192128` and use `16` parallel jobs:
```
make encap-sim PAR_SETS="mceliece6688128 mceliece8192128" NPROC=16
```

# Contact
| Name              | E-Mail                    |
|-------------------|---------------------------|
| Jakub Szefer      | jakub.szefer@yale.edu     |
| Norman Lahr       | norman@lahr.email         |
| Ruben Niederhagen | ruben@polycephaly.org     |
| Sanjay Deshpande  | sanjay.deshpande@yale.edu |
| Tony Chou         | blueprint@crypto.tw       |
| Wen Wang          | wen.wang.ww349@yale.edu   |

# References
[1] Chen, Po-Jen, Tung Chou, Sanjay Deshpande, Norman Lahr, 
Ruben Niederhagen, Jakub Szefer, and Wen Wang. 2022. 
“Complete and Improved FPGA Implementation of Classic McEliece”. 
IACR Transactions on Cryptographic Hardware and Embedded Systems 2022 (3):71-113. 
https://doi.org/10.46586/tches.v2022.i3.71-113

[2] https://eprint.iacr.org/2022/412

[3] https://www.gnu.org/software/make/manual/html_node/Secondary-Expansion.html
