// This file is part of the Classic McEliece hardware source code.
// 
// Copyright (C) 2022
//
// Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
// 
// This source code is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option) any
// later version.
// 
// This source code is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
// 
// You should have received a copy of the GNU General Public License along with
// this source code. If not, see <https://www.gnu.org/licenses/>. 

`timescale 1ns / 1ps

module mceliece_joint_design  #(    
    parameter parameter_set =1,
    
    
    parameter N = (parameter_set == 1)? 3488:
                  (parameter_set == 2)? 4608:
                  (parameter_set == 3)? 6688:
                  (parameter_set == 4)? 6960: 
                                        8192,
                                         
    parameter M = (parameter_set == 1)? 12:13,    
    
    parameter T = (parameter_set == 1)? 64:
                  (parameter_set == 2)? 96:
                  (parameter_set == 3)? 128:
                  (parameter_set == 4)? 119: 
                                        128,                                       
      
    // system parameters
    // key generation
    parameter N_H = T, // N in H elimination T for p1-p4 M for p5
    parameter BLOCK_H = T, // BLOCK in H elimination T for p1-p4 M for p5
    parameter N_R = 2, // N in R elimination
    parameter BLOCK_R = 1, // BLOCK in R elimination

    parameter SIGMA1 = 16, // bit width of random values for generation the polynomial
    parameter SIGMA2 = 32, // bit width of integer used in sort module

    parameter MEM_WIDTH = 32,
    parameter K = N,
    parameter L = M*T,
    parameter K_H_elim = N_H*((K+N_H-1)/N_H),
    parameter K_R_elim = N_R*((T+1+N_R-1)/N_R),
    
    
    parameter S_RAM_DEPTH = (N+(32-N%32)%32)/32,
    parameter SEED_SIZE = 256,
    parameter SEED_RAM_DEPTH = (SEED_SIZE +(32-SEED_SIZE%32)%32)/32,
    
    // ADD FFT parameters
    parameter mem_width = 16, // memory width in add_FFT (mem_A/B)
    parameter mul_sec_BM = 20, // number of multipliers inside vec_mul_multi in BM module
    parameter mul_sec_BM_step = 20, // number of multipliers inside vec_mul_multi in BM module

    parameter eva_mem_WIDTH = 2*M*mem_width, //memory width in the add_FFT module, sum of mem_A and mem_B
    parameter eva_mem_WIDTH_bits = `CLOG2(mem_width)+1,
    parameter eva_mem_DEPTH = (1 << (`CLOG2(T))), //memory depth in the add_FFT module
    parameter eva_mem_DEPTH_bits = `CLOG2(T),
    
    parameter DATA = "",
	parameter KEY_START_ADDR = L*(L/N_H)
    
  )
  ( 
        input 									   clk,
        input 									   rst,
        input [2:0] 							   operation,
        // operation = 000 => KEYGEN
        // operation = 001 => DECAPSULATION
        // operation = 010 => ENCAPSULATION
        
        input [31:0] 							   delta,
        input 									   delta_valid,
        
        
            // S   
        input 									   s_rd_en,
        input [`CLOG2(S_RAM_DEPTH)-1:0] 		   s_rd_addr, 
        output [32-1:0] 						   s_rd_out,
        
        //
        input 									   host_sk_poly_rd_en,
        input [`CLOG2(T*K_R_elim/N_R)-1:0] 		   host_sk_poly_rd_addr,
        output [M*N_R-1:0] 						   sk_poly_dout,
        output 									   sk_poly_gen_done,
        
        // permutation generation
        input 									   host_perm_rd_en,
        input [M-1:0] 							   host_perm_rd_addr,
        output [M-1:0] 							   perm_dout,
        output 									   perm_done,
        
        // PK generation
        input 									   host_pk_rd_en,
        input [`CLOG2(L*K_H_elim/N_H)-1:0] 		   host_pk_rd_addr, 
        output [N_H-1:0] 						   pk_dout,
        output 									   pk_gen_done,
        
        
        output 									   support_gen_done,
            
        input 									   start_decap,
        output 									   done_decap,
        
        input 									   C0_valid,
        input [31:0] 							   C0_in,
        output 									   C0_ready,
        
        input 									   C1_valid, 
        input [31:0] 							   C1_in,
        output 									   C1_ready,
        
        input 									   poly_g_valid, 
        input [31:0] 							   poly_g_in,
        output 									   poly_g_ready, 
                    
        input 									   rd_K,
        input [2:0] 							   K_addr, 
        output [31:0] 							   K_out,
        
       // ENCAPSULATION
        input 									   host_pk_wr_en,
        input [`CLOG2(L*K_H_elim/N_H)-1:0] 		   host_pk_wr_addr, 
		input [N_H-1:0] 						   host_pk_din,
       
        input 									   seed_valid,
        input [31:0] 							   seed,
        output 									   encap_done,
      
        input 									   rd_C0,
        input [`CLOG2((L + (32-L%32)%32)/32) -1:0] C0_addr, 
        output [31:0] 							   C0_out,
    
        input 									   rd_C1,
        input [2:0] 							   C1_addr, 
        output [31:0] 							   C1_out,
    
        input 									   rd_K_encap,
        input [2:0] 							   K_addr_encap, 
        output [31:0] 							   K_out_encap
        
    );
  
 //PARAMETERS FOR INSTRUCTION
 
parameter KEYGEN = 3'b000;
parameter DECAPSULATION  = 3'b001;
parameter ENCAPSULATION  = 3'b010;
 
// FieldOrdering Signals
//KEYGEN
wire                            host_alpha_rand_wr_en_kg; 
wire [M-1:0]                    host_alpha_rand_wr_addr_kg;
wire [SIGMA2-1:0]               host_alpha_rand_data_in_kg;
wire                            sk_gen_start_kg;
wire                            perm_done_kg;
wire                            field_ordering_fail_kg;
wire                            perm_rd_en_valid_kg;
wire [M-1:0]                    perm_rd_addr_valid_kg;
wire [M-1:0]                    perm_dout_kg;
wire [SIGMA2-1:0]               perm_rand_out_kg;  

//DECAP
wire                            host_alpha_rand_wr_en_dec; 
wire [M-1:0]                    host_alpha_rand_wr_addr_dec;
wire [SIGMA2-1:0]               host_alpha_rand_data_in_dec;
wire                            sk_gen_start_dec;
wire                            perm_done_dec;
wire                            field_ordering_fail_dec;
wire                            perm_rd_en_valid_dec;
wire [M-1:0]                    perm_rd_addr_valid_dec;
wire [M-1:0]                    perm_dout_dec;
wire [SIGMA2-1:0]               perm_rand_out_dec; 

wire                            host_alpha_rand_wr_en; 
wire [M-1:0]                    host_alpha_rand_wr_addr;
wire [SIGMA2-1:0]               host_alpha_rand_data_in;
wire                            sk_gen_start;
//wire                            perm_done;
wire                            field_ordering_fail;
wire                            perm_rd_en_valid;
wire [M-1:0]                    perm_rd_addr_valid;
//wire [M-1:0]                    perm_dout;
wire [SIGMA2-1:0]               perm_rand_out;

//shake interface ports   

//KEYGEN
wire                            shake_din_valid_kg;
wire                             shake_din_ready_kg;
wire [31:0]                     shake_din_kg;
wire                             shake_dout_valid_kg;
wire                            shake_dout_ready_kg;
wire [31:0]                      shake_dout_scrambled_kg;
wire                            shake_force_done_kg;



//DECAP     
wire                            shake_din_valid_dec;
wire                            shake_din_ready_dec;
wire [31:0]                     shake_din_dec;
wire                            shake_dout_ready_dec;
wire  [31:0]                     shake_dout_scrambled_dec;
wire                            shake_force_done_dec;
wire                             shake_dout_valid_dec; 


//ENCAP     
wire                            shake_din_valid_enc;
wire                            shake_din_ready_enc;
wire [31:0]                     shake_din_enc;
wire                            shake_dout_ready_enc;
wire  [31:0]                     shake_dout_scrambled_enc;
wire                            shake_force_done_enc;
wire                             shake_dout_valid_enc;

wire                             shake_din_valid;
wire                            shake_din_ready;
wire [31:0]                      shake_din;
wire                             shake_dout_ready;
wire [31:0]                     shake_dout_scrambled;
wire                             shake_force_done;
wire                            shake_dout_valid; 
   
 
 

assign shake_din_valid = (operation == KEYGEN)? shake_din_valid_kg:
                         (operation == DECAPSULATION)? shake_din_valid_dec:
                         (operation == ENCAPSULATION)? shake_din_valid_enc:
                         1'b0;

assign shake_din = (operation == KEYGEN)? shake_din_kg:
                   (operation == DECAPSULATION)? shake_din_dec:
                   (operation == ENCAPSULATION)? shake_din_enc:
                   0;
 

assign shake_force_done = (operation == KEYGEN)? shake_force_done_kg:
                         (operation == DECAPSULATION)? shake_force_done_dec:
                         (operation == ENCAPSULATION)? shake_force_done_enc:
                         1'b0; 

assign shake_dout_ready = (operation == KEYGEN)? shake_dout_ready_kg:
                         (operation == DECAPSULATION)? shake_dout_ready_dec:
                         (operation == ENCAPSULATION)? shake_dout_ready_enc:
                         1'b0;                  
 
assign shake_din_ready_kg = (operation == KEYGEN)? shake_din_ready : 0;
assign shake_din_ready_dec = (operation == DECAPSULATION)? shake_din_ready : 0;
assign shake_din_ready_enc = (operation == ENCAPSULATION)? shake_din_ready : 0;

assign shake_dout_scrambled_kg =shake_dout_scrambled;
assign shake_dout_scrambled_enc =shake_dout_scrambled;
assign shake_dout_scrambled_dec =shake_dout_scrambled;

assign shake_dout_valid_kg = (operation == KEYGEN)? shake_dout_valid : 0;
assign shake_dout_valid_dec = (operation == DECAPSULATION)? shake_dout_valid : 0;
assign shake_dout_valid_enc = (operation == ENCAPSULATION)? shake_dout_valid : 0;
    


 keccak_top shake_instance
 (
    .rst(rst),
    .clk(clk),
    .din_valid(shake_din_valid),
    .din_ready(shake_din_ready),
    .din(shake_din),
    .dout_valid(shake_dout_valid),
    .dout_ready(shake_dout_ready),
    .dout(shake_dout_scrambled),
    .force_done(shake_force_done)
);
     
  seeded_key_gen # (.parameter_set(parameter_set), .M(M), .T(T), .N(N), .N_H(N_H), .BLOCK_H(BLOCK_H), .N_R(N_R), .BLOCK_R(BLOCK_R), .SIGMA1(SIGMA1), .SIGMA2(SIGMA2), .MEM_WIDTH(MEM_WIDTH), .DATA(DATA))
  SEEDEDKEYGEN  (
    .clk(clk),
    .rst(rst),
    .delta_valid((operation == KEYGEN)?delta_valid:0),
    .delta(delta),
    .done(done),
    
    //shake interface
    .shake_din_valid_kg(shake_din_valid_kg),         
    .shake_din_ready_kg(shake_din_ready_kg),          
    .shake_din_kg(shake_din_kg),                 
    .shake_dout_ready_kg(shake_dout_ready_kg),       
    .shake_dout_scrambled_kg(shake_dout_scrambled_kg),      
    .shake_force_done_kg(shake_force_done_kg),          
    .shake_dout_valid_kg(shake_dout_valid_kg),                               
                                                     
    //
    .host_sk_poly_rd_en(host_sk_poly_rd_en),
    .host_sk_poly_rd_addr(host_sk_poly_rd_addr),
    .sk_poly_dout(sk_poly_dout),
    .sk_poly_gen_done(sk_poly_gen_done),

    //
    .host_perm_rd_en(host_perm_rd_en),
    .host_perm_rd_addr(host_perm_rd_addr),
    .perm_dout(perm_dout),
    .perm_done(perm_done),
    .host_pk_rd_en(host_pk_rd_en_internal),
    .host_pk_rd_addr(host_pk_rd_addr_internal),
    .pk_dout(pk_dout_internal),
    .pk_gen_done(pk_gen_done),

    .host_alpha_rand_wr_en_kg(host_alpha_rand_wr_en_kg), 
    .host_alpha_rand_wr_addr_kg(host_alpha_rand_wr_addr_kg),
    .host_alpha_rand_data_in_kg(host_alpha_rand_data_in_kg),
    .sk_gen_start_kg(sk_gen_start_kg),
    .perm_done_kg(perm_done_kg),
    .field_ordering_fail_kg(field_ordering_fail_kg),
    .perm_rd_en_valid_kg(perm_rd_en_valid_kg),
    .perm_rd_addr_valid_kg(perm_rd_addr_valid_kg),
    .perm_dout_kg(perm_dout_kg),
    .perm_rand_out_kg(perm_rand_out_kg),
    
     //shared fft signals
    .fft_eva_start_kg(fft_eva_start_kg),
    .fft_coeff_in_kg(fft_coeff_in_kg),
    .fft_rd_en_kg(fft_rd_en_kg),
    .fft_rd_addr_kg(fft_rd_addr_kg),
    .fft_eva_dout_kg(fft_eva_dout_kg),
    .fft_eva_done_kg(fft_eva_done_kg),
    
    .host_pk_wr_en(host_pk_wr_en),  
    .host_pk_wr_addr(host_pk_wr_addr),
    .host_pk_din(host_pk_din)
    
 );
                    
 
 wire host_pk_rd_en_internal;
 wire [`CLOG2(L*K_H_elim/N_H)-1:0] host_pk_rd_addr_internal;
 wire [N_H-1:0] pk_dout_internal;
 
 wire PK_rd;
 wire [`CLOG2(L*K_H_elim/N_H)-1:0] PK_addr;
 
 // reading out the public key
 
 assign host_pk_rd_en_internal = (operation == ENCAPSULATION)?  PK_rd :
                                 (operation == KEYGEN)?         host_pk_rd_en :
                                                                1'b0;
 
assign host_pk_rd_addr_internal = (operation == ENCAPSULATION)?  PK_addr :
                                  (operation == KEYGEN)?         host_pk_rd_addr :
                                                                 0;

assign pk_dout = pk_dout_internal;

 
 // secret permutation generator
   field_ordering #(.M(M), .SIGMA2(SIGMA2)) perm_gen_inst (
    .clk(clk),
    .start(sk_gen_start),  
    .wr_en(host_alpha_rand_wr_en),  
    .wr_addr(host_alpha_rand_wr_addr),
    .rand_in(host_alpha_rand_data_in),
    .done(perm_done),
    .fail(field_ordering_fail),
    .rd_en(perm_rd_en_valid),
    .rd_addr(perm_rd_addr_valid),
    .index_out(perm_dout),
    .rand_dout(perm_rand_out)
  );
  
//inputs to FieldOrdering
assign host_alpha_rand_wr_en    = (operation == KEYGEN)?        host_alpha_rand_wr_en_kg:
                                  (operation == DECAPSULATION)? host_alpha_rand_wr_en_dec:
                                                                0; 
                                  
assign host_alpha_rand_wr_addr  = (operation == KEYGEN)?        host_alpha_rand_wr_addr_kg:
                                  (operation == DECAPSULATION)? host_alpha_rand_wr_addr_dec[M-1:0]:
                                                                0;
                                                                
assign host_alpha_rand_data_in  = (operation == KEYGEN)?        host_alpha_rand_data_in_kg:
                                  (operation == DECAPSULATION)? host_alpha_rand_data_in_dec:
                                                                0;
                                                                
assign sk_gen_start             = (operation == KEYGEN)?        sk_gen_start_kg:
                                  (operation == DECAPSULATION)? sk_gen_start_dec:
                                                                0;
                                                                
assign perm_rd_en_valid         = (operation == KEYGEN)?        perm_rd_en_valid_kg:
                                  (operation == DECAPSULATION)? perm_rd_en_valid_dec:
                                                                0;
assign perm_rd_addr_valid       = (operation == KEYGEN)?        perm_rd_addr_valid_kg:
                                  (operation == DECAPSULATION)? perm_rd_addr_valid_dec:
                                                                0;

//output from FieldOrdering
assign perm_dout_kg = perm_dout;
assign perm_dout_dec = perm_dout;

assign perm_done_kg = perm_done;
assign perm_done_dec = perm_done;

assign field_ordering_fail_kg = field_ordering_fail;
assign field_ordering_fail_dec = field_ordering_fail;

assign perm_rand_out_kg = perm_rand_out; 
assign perm_rand_out_dec = perm_rand_out; 




  decap_with_support # (.parameter_set(parameter_set), .M(M), .T(T), .N(N), .N_H(N_H), .BLOCK_H(BLOCK_H), .N_R(N_R), .BLOCK_R(BLOCK_R), .SIGMA1(SIGMA1), .SIGMA2(SIGMA2), .MEM_WIDTH(MEM_WIDTH))
  DECAP_WITH_SUPPORT  (
    .clk(clk),
    .rst(rst),
    .delta_valid((operation == DECAPSULATION)?delta_valid:0),
    .delta(delta),
    .start_decap(start_decap),
    .done_decap(done_decap),
    .support_gen_done(support_gen_done),
    
    //shake interface
    .shake_din_valid_d(shake_din_valid_dec),         
    .shake_din_ready_d(shake_din_ready_dec),          
    .shake_din_d(shake_din_dec),                 
    .shake_dout_ready_d(shake_dout_ready_dec),       
    .shake_dout_scrambled_d(shake_dout_scrambled_dec),      
    .shake_force_done_d(shake_force_done_dec),          
    .shake_dout_valid_d(shake_dout_valid_dec),                               
    
    .C0_valid(C0_valid),
    .C0_in(C0_in),
    .C0_ready(C0_ready),
    
    .C1_valid(C1_valid),
    .C1_in(C1_in),
    .C1_ready(C1_ready),
    
    .poly_g_valid(poly_g_valid),
    .poly_g_in(poly_g_in),
    .poly_g_ready(poly_g_ready),

    .rd_K(rd_K),
    .K_addr(K_addr),
    .K_out(K_out),
    
    .host_alpha_rand_wr_en(host_alpha_rand_wr_en_dec), 
    .host_alpha_rand_wr_addr(host_alpha_rand_wr_addr_dec),
    .host_alpha_rand_data_in(host_alpha_rand_data_in_dec),
    .start_fo(sk_gen_start_dec),
    .done_fo(perm_done_dec),
    .field_ordering_fail(field_ordering_fail_dec),
    .P_rd_en(perm_rd_en_valid_dec),
    .P_rd_addr(perm_rd_addr_valid_dec),
    .P_dout(perm_dout_dec),
//    .perm_rand_out(perm_rand_out_dec),
    
    //shared fft signals
    .fft_eva_start_dec(fft_eva_start_dec),
    .fft_coeff_in_dec(fft_coeff_in_dec),
    .fft_rd_en_dec(fft_rd_en_dec),
    .fft_rd_addr_dec(fft_rd_addr_dec),
    .fft_eva_dout_dec(fft_eva_dout_dec),
    .fft_eva_done_dec(fft_eva_done_dec)                                         

 );

wire fft_eva_start_dec;
wire [M*(T+1)-1:0] fft_coeff_in_dec;
wire fft_rd_en_dec;
wire [`CLOG2(T+1)-1:0] fft_rd_addr_dec;
wire [eva_mem_WIDTH-1:0] fft_eva_dout_dec;
wire fft_eva_done_dec;

wire fft_eva_start_kg;
wire [M*(T+1)-1:0] fft_coeff_in_kg;
wire fft_rd_en_kg;
wire [`CLOG2(T+1)-1:0] fft_rd_addr_kg;
wire [eva_mem_WIDTH-1:0] fft_eva_dout_kg;
wire fft_eva_done_kg;

wire fft_eva_start;
wire [M*(T+1)-1:0] fft_coeff_in;
wire fft_rd_en;
wire [`CLOG2(T+1)-1:0] fft_rd_addr;
wire [eva_mem_WIDTH-1:0] fft_eva_dout;
wire fft_eva_done;
  
assign fft_eva_start = (operation == KEYGEN)? fft_eva_start_kg: fft_eva_start_dec;
assign fft_coeff_in = (operation == KEYGEN)? fft_coeff_in_kg: fft_coeff_in_dec;
assign fft_rd_en = (operation == KEYGEN)? fft_rd_en_kg: fft_rd_en_dec;
assign fft_rd_addr = (operation == KEYGEN)? fft_rd_addr_kg: fft_rd_addr_dec;

assign fft_eva_dout_kg = fft_eva_dout;
assign fft_eva_dout_dec = fft_eva_dout;

assign fft_eva_done_kg = (operation == KEYGEN)? fft_eva_done: 1'b0;
assign fft_eva_done_dec = (operation == DECAPSULATION)? fft_eva_done: 1'b0;

   add_FFT #(.numc(T+1), .gf(M), .mem_width(mem_width)) ADD_FFT_SHARED (
    .clk(clk),
    .start(fft_eva_start),
    // coeff input
    .coeff_in(fft_coeff_in),
    // add_FFT output
    .rd_en(fft_rd_en),
    .rd_addr(fft_rd_addr),
    .data_out(fft_eva_dout),
    .done(fft_eva_done)
  );
 

    
    
   encap_seq_gen # (.parameter_set(parameter_set), .m(M), .t(T), .n(N), .e_width(N_H), .col_width(N_H), .KEY_START_ADDR(KEY_START_ADDR))
  ENCAP  (
    .clk(clk),
    .rst(rst),
    .seed_valid(seed_valid),
    .seed(seed),
    
    .PK_addr(PK_addr),
    .PK_col(pk_dout_internal),
    .PK_rd(PK_rd),
    
    .done(encap_done),
    .rd_C0(rd_C0),
    .C0_addr(C0_addr),
    .C0_out(C0_out),
    
    .rd_C1(rd_C1),
    .C1_addr(C1_addr),
    .C1_out(C1_out),
    
    .rd_K(rd_K_encap),
    .K_addr(K_addr_encap),
    .K_out(K_out_encap),

    //shake interface
    .din_valid_shake_enc(shake_din_valid_enc),
    .din_ready_shake(shake_din_ready_enc),
    .din_shake_enc(shake_din_enc),
    .dout_valid_shake(shake_dout_valid_enc),
    .dout_ready_shake_enc(shake_dout_ready_enc),
    .dout_shake(shake_dout_scrambled_enc),
    .force_done_shake(shake_force_done_enc)
  
    );   
  
    
endmodule
