# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

# Joint module with all KEM operations and shared Additive FFT and Field Ordering.

JOINT_DESIGN_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

# Include the parameter set definitions
include $(JOINT_DESIGN_SRC_PATH)/../../config/parameters.mk

# Include toolset
include $(JOINT_DESIGN_SRC_PATH)/../../config/tools.mk

# Name of the design's topmodule
TOPMODULE ?= joint_design

# List of the required submodules
JOINT_DESIGN_SUBMODULES := SEEDEDKEYGEN ENCAPSULATION DECAPSULATION ADDITIVE_FFT FIELD_ORDERING SHAKE256

#
# Directory and path definitions
#
# Build directory
BUILD_DIR ?= $(JOINT_DESIGN_SRC_PATH)/build
# Directory pointing to the known answer tests
KAT_DIR ?= $(BUILD_DIR)/kat
# Directory for the performance results
RESULTS_DIR ?= $(BUILD_DIR)/results

# Source path for the common used modules
COMMON_SRC_PATH ?= $(JOINT_DESIGN_SRC_PATH)/../common
# Source path for the KAT generation script
KATGEN_SRC_PATH ?= $(JOINT_DESIGN_SRC_PATH)/../../kat

.SECONDEXPANSION:

#
# Parameter-specific build subdirectories
#
ifndef BUILD_DIR_TEMPLATE
define BUILD_DIR_TEMPLATE =

# Parameter-specific build directory root
BUILD_DIR_$(1) = $$(BUILD_DIR)/$(1)/$$(TOPMODULE)/$$(SYSTEMIZER)
# Parameter-specific source, testbench, and KAT directory
BUILD_DIR_SRC_$(1) = $$(BUILD_DIR_$(1))/src
BUILD_DIR_TB_$(1) = $$(BUILD_DIR_$(1))/testbench
BUILD_DIR_SYNTH_$(1) = $$(BUILD_DIR_$(1))/synthesis
BUILD_DIR_KAT_$(1) = $$(KAT_DIR/$(1))

endef

$(foreach par, $(PAR_SETS), $(eval $(call BUILD_DIR_TEMPLATE,$(par))))

endif

#
# Configurations and options of the module
#
# The available systemizers are organized in
# common/systemizer/modules.mk:
# SYSTEMIZER_SPEA_REDO
# SYSTEMIZER_SPGF2
# SYSTEMIZER_SPGF2_REDO
# SYSTEMIZER_SPGFX
# SYSTEMIZER_SPGFXA
SYSTEMIZER := SYSTEMIZER_SPEA_REDO

SLICED_PUBKEY_COLUMN_WIDTHS = $(sort $(foreach par,$(PAR_SETS),$$(N_H_elim_$(par))))

#
# Include makefiles for the submodules and KAT generation
#
include $(JOINT_DESIGN_SRC_PATH)/../keygen/module.mk
include $(JOINT_DESIGN_SRC_PATH)/../encap/module.mk
include $(JOINT_DESIGN_SRC_PATH)/../decap/module.mk
include $(COMMON_SRC_PATH)/shake256/module.mk
include $(COMMON_SRC_PATH)/field_ordering/module.mk
include $(COMMON_SRC_PATH)/additive_fft/module.mk
include $(KATGEN_SRC_PATH)/Makefile

#
# Sources definitions
#

# Set the source files of this module
JOINT_DESIGN_SRC = mceliece_kem_combined.v

# Add the sources of the submodules
JOINT_DESIGN_SRC += $(foreach module,$(JOINT_DESIGN_SUBMODULES),$($(module)_SRC))

# Define comulated target for generated sources
.PHONY: sources
sources: $(addprefix sources-, $(PAR_SETS))

# Set of all sources
JOINT_DESIGN =


# Define the parameter-specific sources and targets
define JOINT_DESIGN_SRC_TEMPLATE =

# Define all required targets to be included in the modules above.
# The sort function filters double entries.
JOINT_DESIGN_$(1) = $(sort $(addprefix $$(BUILD_DIR_SRC_$(1))/,$(JOINT_DESIGN_SRC)))

# Collect all targets
JOINT_DESIGN += $$(JOINT_DESIGN_$(1))

# Define parameter-specific sources targets
.PHONY: sources-$(1)
sources-$(1): $$(JOINT_DESIGN_$(1))

# # Make build directory for the sources
# $$(BUILD_DIR_SRC_$(1)):
# 	mkdir -p $$@

#
# Sources generation
#
$$(BUILD_DIR_SRC_$(1))/mceliece_kem_combined.v: $$(JOINT_DESIGN_SRC_PATH)/mceliece_kem_combined.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call JOINT_DESIGN_SRC_TEMPLATE,$(par))))


#
# Testbench sources definitions
#

# Set the testbench sources
JOINT_DESIGN_TB_SRC = mceliece_kem_combined_tb.v
# JOINT_DESIGN_TB_SRC = decap_tb.v

# Define the comulated target for testbench sources generation
.PHONY: testbench
testbench: $(addprefix testbench-, $(PAR_SETS))

# Set of all testbench sources
JOINT_DESIGN_TB =


# Define the parameter-specific testbench targets
define JOINT_DESIGN_TB_TEMPLATE =

# Define all required targets to be included in the modules above.
JOINT_DESIGN_TB_$(1) = $(addprefix $$(BUILD_DIR_TB_$(1))/,$(JOINT_DESIGN_TB_SRC))

# Collect all testbench targets
JOINT_DESIGN_TB += $$(JOINT_DESIGN_TB_$(1))

# Define parameter-specific testbench targets
.PHONY: testbench-$(1)
testbench-$(1): sources-$(1) $$(JOINT_DESIGN_TB_$(1))

# $$(BUILD_DIR_TB_$(1)):
# 	mkdir -p $$@

#
# Sources generation
#
# $$(BUILD_DIR_TB_$(1))/mceliece_kem_combined_tb.v: $$(JOINT_DESIGN_SRC_PATH)/testbench/mceliece_kem_combined_tb.v
#	cp $$< $$@

$$(BUILD_DIR_TB_$(1))/%.v: $$(JOINT_DESIGN_SRC_PATH)/testbench/%.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call JOINT_DESIGN_TB_TEMPLATE,$(par))))

#
# Simulation targets and files
#

# Define Verilator-specific testbench source
JOINT_DESIGN_VERILATOR_TB_SRC = $(JOINT_DESIGN_SRC_PATH)/testbench/tb.cpp

# Define comulated target for the simulation runs und results generation
.PHONY: sim
sim: $(addprefix sim-, $(PAR_SETS))

# Parameter-specific simulator definitions and targets
define JOINT_DESIGN_SIM_TEMPLATE =

# Define public key name for specific slice width
KAT_FILE_PUBKEY_$$(N_H_elim_$(1)) = $(basename $(KAT_FILE_PUBKEY))_$$(N_H_elim_$(1))$(suffix $(KAT_FILE_PUBKEY))

# Required KAT files for joint design
JOINT_DESIGN_KAT_$(1) = \
	$$(addprefix $$(KAT_DIR_$(1))/, \
	$(KAT_FILE_POLY_G))
	# $(KAT_FILE_CIPHER_0) \
	# $(KAT_FILE_CIPHER_1) \
	# $(KAT_FILE_DELTA_PRIME) \
	# $(KAT_FILE_K) \
	# $(KAT_FILE_P) \

# Set of all files to perform the simulation
JOINT_DESIGN_SIM_$(1) = \
	$$(JOINT_DESIGN_VERILATOR_TB_SRC) \
	$$(JOINT_DESIGN_TB_$(1)) \
	$$(JOINT_DESIGN_$(1)) \
	$$(JOINT_DESIGN_KAT_$(1))

# Simulation binaries
JOINT_DESIGN_VERILATOR_BIN_$(1) = $$(BUILD_DIR_TB_$(1))/mceliece_kem_combined_tb.bin

# Simulation performance files
JOINT_DESIGN_RESULT_TB_$(1) = $$(RESULTS_DIR)/cycles_profile_$(1)_$(TOPMODULE)_$(SYSTEMIZER).data

# Define parameter-specific simulation results
SIM_$(1) ?=
SIM_$(1) += $$(JOINT_DESIGN_RESULT_TB_$(1))

# Simulation binary target
#	-Wall -Wno-fatal
$$(JOINT_DESIGN_VERILATOR_BIN_$(1)): $$(JOINT_DESIGN_SIM_$(1)) | $$(RESULTS_DIR)
	$(VERILATOR) \
	-Wno-lint -Wno-fatal -Wno-style \
	--timescale-override 1ps/1ps \
	--unroll-count 1024 \
	--trace \
	--Mdir $$(BUILD_DIR_TB_$(1))/obj_dir \
	--top-module mceliece_kem_combined_tb \
	-DSHARED \
	-Gparameter_set=$$(id_$(1)) \
	-Gcol_width=$$(N_H_elim_$(1)) \
	-Ge_width=$$(N_H_elim_$(1)) \
	-GKEY_START_ADDR=0 \
	-GN_H=$$(N_H_elim_$(1)) \
	-GBLOCK_H=$$(BLOCK_H_elim_$(1)) \
	-GN_R=$$(N_R_elim) \
	-GBLOCK_R=$$(BLOCK_R_elim) \
	-GSIGMA1=$$(SIGMA1) \
	-GSIGMA2=$$(SIGMA2) \
	-GMEM_WIDTH=$$(mem_width_keygen_$(1)) \
	-DFILE_C0=\"$$(KAT_DIR_$(1))/$(KAT_FILE_CIPHER_0)\" \
	-DFILE_C1=\"$$(KAT_DIR_$(1))/$(KAT_FILE_CIPHER_1)\" \
	-DFILE_CIPHER0_OUT=\"$$(BUILD_DIR_TB_$(1))/cipher_0.out\" \
	-DFILE_CIPHER1_OUT=\"$$(BUILD_DIR_TB_$(1))/cipher_1.out\" \
	-DFILE_CYCLES_PROFILE=\"$$(JOINT_DESIGN_RESULT_TB_$(1))\" \
	-DFILE_DELTA=\"$$(KAT_DIR_$(1))/$(KAT_FILE_DELTA_PRIME)\" \
	-DFILE_DOUBLED_SYNDROME_OUT=\"$$(BUILD_DIR_TB_$(1))/doubled_syndrome.out\" \
	-DFILE_ERROR_DECAP_OUT=\"$$(BUILD_DIR_TB_$(1))/error_decap.out\" \
	-DFILE_ERROR_ENCAP_OUT=\"$$(BUILD_DIR_TB_$(1))/error_encap.out\" \
	-DFILE_K_ENCAP_OUT=\"$$(BUILD_DIR_TB_$(1))/K_encap.out\" \
	-DFILE_K_DECAP_OUT=\"$$(BUILD_DIR_TB_$(1))/K_decap.out\" \
	-DFILE_MEM_CONSTS=\"$$(BUILD_DIR_SRC_$(1))/mem_consts\" \
	-DFILE_MEM_SEED=\"$$(KAT_DIR_$(1))/$(KAT_FILE_SEED)\" \
	-DFILE_P=\"$$(KAT_DIR_$(1))/$(KAT_FILE_P)\" \
	-DFILE_PK_OUT=\"$$(BUILD_DIR_TB_$(1))/pubkey.out\" \
	-DFILE_PK_SLICED=\"$$(KAT_DIR_$(1))/$$(KAT_FILE_PUBKEY_$$(N_H_elim_$(1)))\" \
	-DFILE_POLYG=\"$$(KAT_DIR_$(1))/$(KAT_FILE_POLY_G)\" \
	-DFILE_RE_ENC_OUT=\"$$(BUILD_DIR_TB_$(1))/re_encryption_res.out\" \
	-DFILE_SEED=\"$$(KAT_DIR_$(1))/$$(KAT_FILE_DELTA_PRIME)\" \
	-DFILE_VCD=\"$$(BUILD_DIR_TB_$(1))/wave.vcd\" \
	-I$$(BUILD_DIR_SRC_$(1)) \
	-I$$(BUILD_DIR_TB_$(1)) \
	--cc \
	$$(BUILD_DIR_SRC_$(1))/clog2.v \
	$$(JOINT_DESIGN_TB_$(1)) \
	$$(filter-out $$(BUILD_DIR_SRC_$(1))/mem_consts $$(BUILD_DIR_SRC_$(1))/clog2.v,$$(JOINT_DESIGN_$(1))) \
	--exe $$< \
	--build \
	-j 8 \
	--threads 1 \
	--x-initial 0 \
	-o $$@

# Perform the simulation, record the performance data, and check the outcome.
$$(JOINT_DESIGN_RESULT_TB_$(1)): $$(JOINT_DESIGN_VERILATOR_BIN_$(1)) $$(JOINT_DESIGN_KAT_$(1))
	$$<
	@diff $$(KAT_DIR_$(1))/$(KAT_FILE_PUBKEY) $$(BUILD_DIR_TB_$(1))/pubkey.out \
	&& echo "[$(1)-seededkeygen] Test Passed!"
	@$(SAGE) $(MCELIECE_KATGEN) \
	--check-encaps \
	--kat-path $$(KAT_DIR_$(1)) \
	--pars $(1) \
	--sessionkey $$(BUILD_DIR_TB_$(1))/K_encap.out \
	--cipher $$(BUILD_DIR_TB_$(1))/cipher_0_$(2).out $$(BUILD_DIR_TB_$(1))/cipher_1.out \
	&& echo "[$(1)-encapsulation] Test Passed!"
	@sed -i -e ':a' -e 'N;s/\n//;ba' $$(BUILD_DIR_TB_$(1))/K_decap.out 
	@diff $$(BUILD_DIR_TB_$(1))/K_decap.out $$(KAT_DIR_$(1))/$(KAT_FILE_K) \
	&& echo "[$(1)-decapsulation] Test Passed!"


# Define parameter-specific generated testbench run targets
.PHONY: sim-$(1)
sim-$(1): $$(SIM_$(1))

endef

$(foreach par, $(PAR_SETS), \
	$(eval $(call JOINT_DESIGN_SIM_TEMPLATE,$(par))))

# Define comulated target for generated sources
.PHONY: sim
sim: $(addprefix sim-, $(PAR_SETS))


#
# Synthesis targets and files
#
# Include vendor- and devices-specific definitions and recipes
include $(JOINT_DESIGN_SRC_PATH)/../../config/synthesis.mk

# Define the synthesis recipes for Xilinx devices
define JOINT_DESIGN_SYNTH_TEMPLATE =

# Define required report files
JOINT_DESIGN_RESULT_SYNTH_UTILIZATION_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_utilization_$(1)_$(2).txt
JOINT_DESIGN_RESULT_SYNTH_UTILIZATION_HIERARCHICAL_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_utilization_hierarchical_$(1)_$(2).txt
JOINT_DESIGN_RESULT_SYNTH_TIMING_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_timing_$(1)_$(2).txt
JOINT_DESIGN_RESULT_SYNTH_TIMING_SUMMARY_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_timing_summary_$(1)_$(2).txt
JOINT_DESIGN_RESULT_SYNTH_CLOCKS_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_clocks_$(1)_$(2).txt

# Perform the synthesis
# #TODO (MAKE versions < 3.8 do not support mutli-target rules)
$$(JOINT_DESIGN_RESULT_SYNTH_CLOCKS_$(1)_$(2)): $$(JOINT_DESIGN_$(2)) $$(XILINX_RESOURCE_ANALYSIS_TCL) $$(XILINX_CONSTRAINTS_TIMING_$(1)) | $(RESULTS_DIR)
	$$(VIVADO) \
	-nojournal \
	-log $$(RESULTS_DIR)/$(TOPMODULE)_synthesis_$(1)_$(2).log \
	-mode batch \
	-source $$(XILINX_RESOURCE_ANALYSIS_TCL) \
	-tclargs \
		"$$(BUILD_DIR_SRC_$(2))/clog2.v $(filter-out $$(BUILD_DIR_SRC_$(2))/clog2.v,$$(JOINT_DESIGN_$(2)))" \
		$$(XILINX_CONSTRAINTS_TIMING_$(1)) \
		decap_with_support \
		$$(XILINX_FPGA_MODEL_ID_$(1)) \
		$$(JOINT_DESIGN_RESULT_SYNTH_UTILIZATION_$(1)_$(2)) \
		$$(JOINT_DESIGN_RESULT_SYNTH_UTILIZATION_HIERARCHICAL_$(1)_$(2)) \
		$$(JOINT_DESIGN_RESULT_SYNTH_TIMING_$(1)_$(2)) \
		$$(JOINT_DESIGN_RESULT_SYNTH_TIMING_SUMMARY_$(1)_$(2)) \
		$$(JOINT_DESIGN_RESULT_SYNTH_CLOCKS_$(1)_$(2)) \
		"parameter_set=$$(id_$(2)) col_width=$$(N_H_elim_$(2)) e_width=$$(N_H_elim_$(2)) " \
		"KEY_START_ADDR=0 N_H=$$(N_H_elim_$(2)) BLOCK_H=$$(BLOCK_H_elim_$(2)) " \
		"N_R=$$(N_R_elim) BLOCK_R=$$(BLOCK_R_elim) SIGMA1=$$(SIGMA1) SIGMA2=$$(SIGMA2) " \
		"MEM_WIDTH=$$(mem_width_keygen_$(2)) BANKS=$$(BANKS) SRAM_DELAY=$$(SRAM_DELAY) "\
		FILE_MEM_CONSTS=$$(BUILD_DIR_SRC_$(2))/mem_consts SHARED


.PHONY: synthesis-$(1)-$(2)

synthesis-$(1)-$(2): $$(JOINT_DESIGN_RESULT_SYNTH_CLOCKS_$(1)_$(2))

# Define parameter-specific synthesis results
SYNTH_$(2) ?=
SYNTH_$(2) += $$(JOINT_DESIGN_RESULT_SYNTH_CLOCKS_$(1)_$(2))

endef

$(foreach model, $(XILINX_MODELS), \
	$(foreach par, $(PAR_SETS), \
		$(eval $(call JOINT_DESIGN_SYNTH_TEMPLATE,$(model),$(par))) \
	) \
)

# Define synthesis and results targets
define JOINT_DESIGN_RESULTS_TEMPLATE =

.PHONY: synthesis-$(1)
synthesis-$(1): $$(SYNTH_$(1))

.PHONY: results-$(1)
results-$(1): $$(SIM_$(1)) $$(SYNTH_$(1))

endef

$(foreach par, $(PAR_SETS), $(eval $(call JOINT_DESIGN_RESULTS_TEMPLATE,$(par))))


# Define comulated target for the synthesis
.PHONY: synthesis
synthesis: $(addprefix synthesis-, $(PAR_SETS))

# Define comulated target for all performance results
.PHONY: results
results: $(addprefix results-, $(PAR_SETS))

# Add the build directory as order-only-prerequisites to every source file target
$(JOINT_DESIGN) $(JOINT_DESIGN_TB): | $$(@D)

$(RESULTS_DIR):
	mkdir -p $@

clean-joint:
	rm -rf $(BUILD_DIR)
