/*
 * Copyright (C) 2021
 * Author: Sanjay Deshpande <sanjay.deshpande@yale.edu>
 *         Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

`timescale 1ns/1ps

module mceliece_kem_combined_tb
       #(
           parameter CLK_PERIOD = 2,
           parameter parameter_set =1,

           parameter N = (parameter_set == 1)? 3488:
           (parameter_set == 2)? 4608:
           (parameter_set == 3)? 6688:
           (parameter_set == 4)? 6960:
           8192,

           parameter M = (parameter_set == 1)? 12:13,

           parameter T = (parameter_set == 1)? 64:
           (parameter_set == 2)? 96:
           (parameter_set == 3)? 128:
           (parameter_set == 4)? 119:
           128,

           // system parameters
           // key generation
           parameter N_H = T, // N in H elimination T for p1-p4 M for p5
           parameter BLOCK_H = T, // BLOCK in H elimination T for p1-p4 M for p5
           parameter N_R = 2, // N in R elimination
           parameter BLOCK_R = 1, // BLOCK in R elimination

           parameter SIGMA1 = 16, // bit width of random values for generation the polynomial
           parameter SIGMA2 = 32, // bit width of integer used in sort module

           parameter MEM_WIDTH = 32,
           parameter K = N,
           parameter PG_WIDTH = M*(T+1),

           parameter L = M*T,
           parameter K_H_elim = N_H*((N+N_H-1)/N_H),
           parameter K_R_elim = N_R*((T+1+N_R-1)/N_R),

           parameter col_width = 64,
           parameter e_width = 64, // error_width

           parameter n_elim = col_width*((N+col_width-1)/col_width),
           parameter l_n_elim = L*n_elim,
           parameter KEY_START_ADDR = L*(L/col_width),

           parameter MODULE_NAME = "Joint Design",
           parameter MODULE_NAME_K = "Seeded Keygen",
           parameter MODULE_NAME_E = "Encapsulation",
           parameter MODULE_NAME_D = "Decapsulation"
       )
       (
           input wire 	   clk,
           input reg 	   rst,
           input reg [2:0] operation,
           output reg 	   loading_done = 1'b0,
           //output wire 	   done,
           output wire 	   done_keygen,
           output wire 	   done_encap,
           output wire 	   done_decap,
           output wire 	   support_gen_done
       );

reg delta_valid = 1'b0;
reg  [32-1:0] delta;

/* Keygen signals */
// ram controls
reg [4:0] addr = 0;
wire [32-1:0] seed_from_ram_keygen;

reg host_sk_poly_rd_en = 1'b0;
reg [`CLOG2(T*K_R_elim/N_R)-1:0]   host_sk_poly_rd_addr = 0;
wire [M*N_R-1:0] sk_poly_dout;
wire sk_poly_gen_done;

reg host_perm_rd_en = 1'b0;
reg [M-1:0] host_perm_rd_addr = 0;
wire [M-1:0] perm_dout;
wire perm_done;

reg host_pk_rd_en = 0;
reg [`CLOG2(L*K_H_elim/N_H)-1:0] host_pk_rd_addr;
wire [N_H-1:0] pk_dout;
reg pk_gen_fail = 0;

/* Decapsulation signals */
//ram controls
reg [4:0] addr_seed;
wire [32-1:0] seed_from_ram_decap;
wire [32-1:0] seed_from_ram_encap;

reg [`CLOG2((L + (32-L%32)%32)/32) -1 : 0] addr_C0 = 0;
wire [31:0] 								  C0_in;
reg 										  C0_valid = 0;

wire 									  C1_ready;
reg 										  C1_valid;
wire [31:0] 								  C1_in;
reg [2:0] 								  addr_C1 = 0;

wire [31:0] 								  poly_g_in;
reg 										  poly_g_valid;
wire 									  poly_g_ready;
reg [`CLOG2((PG_WIDTH + (32-PG_WIDTH%32)%32)/32) -1 : 0] addr_pg = 0;

reg 										  rd_K;
reg [2:0] 								  K_addr;
wire [31:0] 								  K_out;

reg 										  start_decap;
//wire 									  done_decap;
reg 										  s_valid = 1'b0;
reg [4:0] 								  addr_s = 0;

reg [32-1:0] 						  seed;
reg 										  seed_valid = 1'b0;
reg 										  K_col_valid = 1'b0;
wire [col_width - 1:0] 					  PK_col;

wire done_keygen_int;

mceliece_joint_design # (
                          .parameter_set(parameter_set),
                          .M(M), .T(T), .N(K),
                          .N_H(N_H), .BLOCK_H(BLOCK_H),
                          .N_R(N_R), .BLOCK_R(BLOCK_R),
                          .SIGMA1(SIGMA1), .SIGMA2(SIGMA2),
                          .MEM_WIDTH(MEM_WIDTH),
						  .DATA(`FILE_PK_SLICED),
						  .KEY_START_ADDR(KEY_START_ADDR)
                      )
                      DUT (
                          .clk(clk),
                          .rst(rst),
                          .operation(operation),
                          .delta_valid(delta_valid || s_valid),
                          .delta(delta),

                          // SEEDED KEYGEN PORTS
                          .host_sk_poly_rd_en(host_sk_poly_rd_en),
                          .host_sk_poly_rd_addr(host_sk_poly_rd_addr),
                          .sk_poly_dout(sk_poly_dout),
                          .sk_poly_gen_done(sk_poly_gen_done),
                          .host_perm_rd_en(host_perm_rd_en),
                          .host_perm_rd_addr(host_perm_rd_addr),
                          .perm_dout(perm_dout),
                          .perm_done(perm_done),
                          .host_pk_rd_en(host_pk_rd_en),
                          .host_pk_rd_addr(host_pk_rd_addr),
                          .pk_dout(pk_dout),
                          .pk_gen_done(done_keygen_int),

                          .start_decap(start_decap),
                          .done_decap(done_decap),
                          .support_gen_done(support_gen_done),

                          .C0_valid(C0_valid),
                          .C0_in(C0_in),
                          .C0_ready(C0_ready),

                          .C1_valid(C1_valid),
                          .C1_in(C1_in),
                          .C1_ready(C1_ready),

                          .poly_g_valid(poly_g_valid),
                          .poly_g_in(poly_g_in),
                          .poly_g_ready(poly_g_ready),

                          .rd_K(rd_K),
                          .K_addr(K_addr),
                          .K_out(K_out),

                          .seed_valid(seed_valid),
                          .seed(seed),

                          .encap_done(done_encap),
                          .rd_C0(rd_C0),
                          .C0_addr(C0_addr),
                          .C0_out(C0_out),

                          .rd_C1(rd_C1),
                          .C1_addr(C1_addr),
                          .C1_out(C1_out),

                          .rd_K_encap(rd_K),
                          .K_addr_encap(K_addr),
                          .K_out_encap(K_out),

                          .host_pk_wr_en(0),
                          .host_pk_wr_addr(0),
                          .host_pk_din(0)
                      );

// Dump waveform.
// Caution: the seed keygen requires lots of cycles which results in a
// vcd-file approx. greater than 50GB.
// initial
//  begin
//    $dumpfile(`FILE_VCD);
//    $dumpvars(0, seeded_key_gen_tb);
//  end

/* Clock cycle count profiling */
integer f_cycles_profile;
//   time    time_decap_start;
time    time_sk_gen_start;
time    time_pk_gen_start;
integer time_pk_gen_right_start = -1;
reg [20*8+17*8-1:0] prefix = 0;

// Counter to control the test procedure
reg [31:0] ctr = 0;
integer SIZE_C0 = (L + (32-L%32)%32)/32;
integer SIZE_C1 = 8;
integer SIZE_K = 8;
integer SIZE_PK = K*L/col_width;
integer SIZE_POLY = ((PG_WIDTH + (32-PG_WIDTH%32)%32)/32);
integer SIZE_S = 8;
integer SIZE_SEED = 16;
integer SIZE_TOTAL_DECAP = STOP_POLY;
integer SIZE_TOTAL_ENCAP = STOP_SEED;
integer START_C0 = STOP_C1;
integer START_C1 = STOP_S + 1;
integer START_PK = STOP_SEED + 1;
integer START_POLY = STOP_C0;
integer START_S = 0;
integer START_SEED = 0;
integer STOP_C0 = START_C0 + SIZE_C0;
integer STOP_C1 = START_C1 + SIZE_C1;
integer STOP_PK = START_PK + SIZE_PK;
integer STOP_POLY = START_POLY + SIZE_POLY + 1;
integer STOP_S = SIZE_S + 2;
integer STOP_SEED = START_SEED + SIZE_SEED;

// Set control signals and data
always @(posedge clk)
begin
    if (rst)
    begin
        C0_valid <= 1'b0;
        C1_valid <= 1'b0;
        K_col_valid <= 1'b0;
        addr_C0 <= 0;
        addr_C1 <= 0;
        addr_pg <= 0;
        addr_s <= 0;
        addr_seed <= 0;
        ctr <= 0;
        loading_done <= 1'b0;
        poly_g_valid <= 1'b0;
        s_valid <= 1'b0;
        delta_valid <= 1'b0;
        seed_valid <= 1'b0;
    end // if (rst)
    else
    case(operation)
        3'b000: // seeded keygen
        begin
            ctr <= ctr < 40 ? ctr + 1 : ctr;
            delta <= seed_from_ram_keygen;

            // Time delta_valid with output from memory:
            addr <= (ctr > 21) && (ctr < 29) ? addr + 1 : addr;
            delta_valid <= (ctr > 22) && (ctr < 31) ? 1'b1 : 1'b0;
        end
        3'b001: // decapsulation
        begin
            // Wait until the field ordering module is done.
            if(ctr >= SIZE_TOTAL_DECAP+1 || ctr == STOP_S && support_gen_done == 1'b0)
            begin
                ctr <= ctr;
            end
            else
                // Count until all memory is written.
                if (ctr < SIZE_TOTAL_DECAP+1)
                begin
                    ctr <= ctr + 1;
                end
        end
        3'b010: // encapsulation
        begin
            // Wait until the field ordering module is done.
            if(ctr >= SIZE_TOTAL_ENCAP+1)// || ctr == STOP_SEED && PK_ready == 1'b0)
            begin
                ctr <= ctr;
            end
            else
                // Count until all memory is written.
                if (ctr < SIZE_TOTAL_ENCAP+1)
                begin
                    ctr <= ctr + 1;
                end
        end
        default:
            $display("Operating mode not supported!\n");
    endcase
end


always @(posedge clk)
begin
    case(operation)
        3'b000: // seeded keygen
        begin
        end
        3'b001: // decapsulation
        begin
            delta <= seed_from_ram_decap;
            start_decap <= 1'b0;

            // loading s
            if (ctr >= START_S && ctr < STOP_S)
            begin
                addr_s <= addr_s + 1;
                if (ctr > START_S)
                    s_valid <= 1'b1;
            end
            else
            begin
                addr_s <= addr_s;
                s_valid <= 1'b0;
            end
            // loading C1
            if (ctr > START_C1 && ctr <= STOP_C1)
            begin
                addr_C1 <= addr_C1 + 1;
                C1_valid <= 1'b1;
            end
            else
            begin
                addr_C1 <= addr_C1;
                C1_valid <= 1'b0;
            end
            // loading C0
            if (ctr > START_C0 && ctr <= STOP_C0)
            begin
                addr_C0 <= addr_C0 + 1;
                C0_valid <= 1'b1;
            end
            else
            begin
                addr_C0 <= addr_C0;
                C0_valid <= 1'b0;
            end
            // loading g
            if (ctr > START_POLY && ctr < STOP_POLY)
            begin
                addr_pg <= addr_pg + 1;
                poly_g_valid <= 1'b1;
            end
            else
            begin
                addr_pg <= addr_pg;
                poly_g_valid <= 1'b0;
            end
            // start decapsulation
            if (ctr == SIZE_TOTAL_DECAP)
            begin
                start_decap <= 1'b1;
            end
        end

        3'b010: // encapsulation
        begin
            seed <= seed_from_ram_encap;

            // loading seed
            if (ctr >= START_SEED && ctr < STOP_SEED)
            begin
                seed_valid <= 1'b1;
                if (ctr > START_SEED)
                    addr_seed <= addr_seed + 1;
            end
            else
            begin
                addr_seed <= addr_seed;
                seed_valid <= 1'b0;
            end
        end
    endcase
end




initial
begin
    f_cycles_profile = $fopen(`FILE_CYCLES_PROFILE,"w");
    $sformat(prefix, "[%s] [mceliece%0d%0d]", MODULE_NAME, DUT.N, DUT.T);
    //operation <= 3'b000;
end

/* Seeded Keygen Testbench */
always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.sk_gen_start)
begin
    $display("%s %s started.", prefix, MODULE_NAME_K);
    time_sk_gen_start = $time;
    $fwrite(f_cycles_profile, "%s %0d\n", MODULE_NAME_K, $time/CLK_PERIOD);
    $fflush();
end

always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.sk_poly_gen_done)
begin
    $display("%s %s Generation of g(x) finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "sk_poly_gen %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();
end

// always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.perm_done)
// begin
//     $display("%s %s Permutation finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_sk_gen_start)/CLK_PERIOD);
//     $fwrite(f_cycles_profile, "permutation %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
//     $fflush();
// end

always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.pk_gen_start)
begin
    time_pk_gen_start = $time;
    $display("%s %s Generation of the secret key finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "sk_gen %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();
end

always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.pk_gen_inst.success or posedge DUT.SEEDEDKEYGEN.KEYGEN_module.pk_gen_inst.fail)
begin
    if(ctr > 10)
    begin
        time_pk_gen_right_start = $time;
        $display("%s %s Checking left part finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_pk_gen_start)/CLK_PERIOD);
        $fwrite(f_cycles_profile, "check_left %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
        $fflush();
    end
end

always @(posedge DUT.SEEDEDKEYGEN.KEYGEN_module.pk_gen_inst.elim_done)
begin
    if ((time_pk_gen_right_start > 1) && DUT.SEEDEDKEYGEN.KEYGEN_module.pk_gen_inst.success)
    begin
        $display("%s %s Computing right part finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_pk_gen_start)/CLK_PERIOD);
        $fwrite(f_cycles_profile, "compute_right %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
        $fflush();
    end
end

reg done = 1'b0;

always  @(posedge clk)
  begin
    done <= host_pk_rd_addr == L*K_H_elim/N_H-1;
  end

assign done_keygen = done;

integer f2, i, j, k;

reg read_data = 1'b0;
reg read_data_del = 1'b0;

always @(posedge done_keygen_int)
  begin
    $display("%s %s Generation of public key finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_pk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "pk_gen %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
    $fflush();

    $display("%s %s Generation of key pair finished. (%0d cycles)", prefix, MODULE_NAME_K, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "total %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();

    $display("Writing key to file %s.", `FILE_PK_OUT);

    f2 = $fopen(`FILE_PK_OUT,"w");

    read_data <= 1'b1;
    host_pk_rd_en <= 1'b1;
end

reg [`CLOG2(L*K_H_elim/N_H)-1:0] host_pk_rd_addr_del;

always 
  @(posedge clk)
    begin
      host_pk_rd_addr_del <= host_pk_rd_addr;
      read_data_del <= read_data;

      host_pk_rd_addr <= rst || done ? L*(L/N_H) :
                         read_data ?
                         ( host_pk_rd_addr + L < L*(K_H_elim/N_H) ? host_pk_rd_addr + L :
                           L*(L/N_H) + (host_pk_rd_addr % L) + 1)
                         : host_pk_rd_addr;

      if (read_data_del == 1'b1)
      begin
        if (host_pk_rd_addr_del < L*(K_H_elim/N_H) - L)
          begin
            for (j = 0; j < N_H; j = j+1)
              $fwrite(f2, "%b", pk_dout[j]);
          end
        else
          begin
            for (j = 0; j < N_H - (K_H_elim - K); j = j+1)
              $fwrite(f2, "%b", pk_dout[j]);
            $fwrite(f2, "\n");
          end
      end

      if (done)
        $fclose(f2);
    end

mem_single #(.WIDTH(32), .DEPTH(8), .FILE(`FILE_SEED) ) mem_init_seed_keygen
           (
               .clock(clk),
               .data(0),
               .address(addr),
               .wr_en(0),
               .q(seed_from_ram_keygen)
           );

/* Decapsulation Testbench */
time time_decap_start = 0;
time time_field_ordering_start;
time time_field_ordering;

always @(posedge DUT.perm_gen_inst.start)
begin
   time_field_ordering_start = $time;
   $display("%s FieldOrdering module started at cycle %0d.", prefix, time_field_ordering_start/CLK_PERIOD);
   $fflush();
end

always @(posedge DUT.perm_gen_inst.done)
begin
   time_field_ordering = ($time-time_field_ordering_start)/CLK_PERIOD;
   $display("%s FieldOrdering finished. (%0d cycles)", prefix, time_field_ordering);
   $fwrite(f_cycles_profile, "field_ordering %0d %0d\n", time_field_ordering_start/CLK_PERIOD, time_field_ordering);
   $fflush();
end

always @(posedge start_decap)
begin
   time_decap_start = $time;
   $display("%s %s module started at cycle %0d.", prefix, MODULE_NAME_D, time_decap_start/CLK_PERIOD);
   $fflush();
end

always @(posedge done_decap)
begin
    $display("%s %s module finished. (%0d cycles) [re-encrypt %s]", prefix, MODULE_NAME_D, ($time-time_decap_start)/CLK_PERIOD, DUT.DECAP_WITH_SUPPORT.DECAPSULATION.decryption_module.decryption_fail ? "failed" : "succeeded");
    $display("%s %s FieldOrdering and Decapsulation require %0d cycles.", prefix, MODULE_NAME_D,time_field_ordering+($time-time_decap_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "decap %0d %0d\n", time_decap_start/CLK_PERIOD, ($time-time_decap_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "total %0d %0d\n", time_field_ordering_start/CLK_PERIOD, time_field_ordering+($time-time_decap_start)/CLK_PERIOD);
    $fflush();
end

// output file for K
always @(posedge DUT.DECAP_WITH_SUPPORT.done_decap)
begin
    $writememb(`FILE_K_DECAP_OUT, DUT.DECAP_WITH_SUPPORT.DECAPSULATION.hash_mem.mem,0,7);
    $fflush();
end

mem_single #(.WIDTH(32), .DEPTH(8), .FILE(`FILE_DELTA) ) mem_init_seed_decap
           (
               .clock(clk),
               .data(0),
               .address(addr_s),
               .wr_en(0),
               .q(seed_from_ram_decap)
           );

mem_single #(.WIDTH(32), .DEPTH(8), .FILE(`FILE_C1)) C1_input_mem
           (
               .clock(clk),
               .data(0),
               .address(addr_C1),
               .wr_en(0),
               .q(C1_in)
           );

mem_single #(.WIDTH(32), .DEPTH(((L + (32-L%32)%32)/32)), .FILE(`FILE_C0) ) C0_input_mem
           (
               .clock(clk),
               .data(0),
               .address(addr_C0),
               .wr_en(0),
               .q(C0_in)
           );

mem_single #(.WIDTH(32), .DEPTH(((PG_WIDTH + (32-PG_WIDTH%32)%32)/32)), .FILE(`FILE_POLYG) ) pg_input_mem
           (
               .clock(clk),
               .data(0),
               .address(addr_pg),
               .wr_en(0),
               .q(poly_g_in)
           );


/* Encapsulation Testbench */
time time_encap_start;
time time_encapsulation;
time time_encrypt_start;
time time_encrypt;
time time_fixedweight_start;
time time_fixedweight;


always @(posedge DUT.ENCAP.done)
begin
    $writememb(`FILE_K_ENCAP_OUT, DUT.ENCAP.hash_mem.mem,0,7);
    $writememb(`FILE_CIPHER0_OUT, DUT.ENCAP.encryption_unit.encrypt_mem.mem);
    $writememb(`FILE_CIPHER1_OUT, DUT.ENCAP.C1_mem.mem);
    $writememb(`FILE_ERROR_ENCAP_OUT, DUT.ENCAP.error_vector_gen.onegen_instance.mem_dual_B.mem);
    $fflush();
end

always @(posedge seed_valid)
begin
   time_encap_start = $time;
   $display("%s %s started at cycle %0d.", prefix, MODULE_NAME_E, time_encap_start/CLK_PERIOD);
   $fwrite(f_cycles_profile, "%s %0d\n", MODULE_NAME_E, time_encap_start);
   $fflush();
end

always @(posedge done_encap)
begin
    time_encapsulation = ($time-time_encap_start)/CLK_PERIOD;
    $display("%s %s finished. (%0d cycles).", prefix, MODULE_NAME_E, time_encapsulation);
    $fwrite(f_cycles_profile, "total %0d %0d\n", time_encap_start/CLK_PERIOD, time_encapsulation);
    $fflush();
end

always @(negedge rst)
begin
    if (operation == 3'b010)
    begin
       time_fixedweight_start = $time-time_encap_start;
       $display("%s %s Start FixedWeight at cycle %0d.", prefix, MODULE_NAME_E, time_fixedweight_start/CLK_PERIOD);
       $fflush();
    end
end

always @(posedge DUT.ENCAP.done_error)
begin
    time_fixedweight = ($time-time_fixedweight_start)/CLK_PERIOD;
    $display("%s %s FixedWeight finished. (%0d cycles)", prefix, MODULE_NAME_E, time_fixedweight);
    $fwrite(f_cycles_profile, "fixedweight %0d %0d\n", time_fixedweight_start/CLK_PERIOD, time_fixedweight);
    $fflush();
end

always @(posedge DUT.ENCAP.done_error)
begin
    time_encrypt_start = $time;
    $display("%s %s Start Encode at cycle %0d.", prefix, MODULE_NAME_E, time_encrypt_start/CLK_PERIOD);
    $fflush();
end

always @(posedge DUT.ENCAP.done_encrypt)
begin
    time_encrypt = ($time-time_encrypt_start)/CLK_PERIOD;
    $display("%s %s Encode finished. (%0d cycles)", prefix, MODULE_NAME_E, time_encrypt);
    $fwrite(f_cycles_profile, "encode %0d %0d\n", time_encrypt_start/CLK_PERIOD, time_encrypt);
    $fflush();
end

mem_single #(.WIDTH(32), .DEPTH(16), .FILE(`FILE_MEM_SEED) ) mem_init_seed_encap
           (
               .clock(clk),
               .data(0),
               .address(addr_seed),
               .wr_en(0),
               .q(seed_from_ram_encap)
           );

endmodule
