/*
 * function: this module computes the H matrix.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

module gen_H
#(
  parameter N = 20,
  parameter K = 6960,
  parameter M_elim = 1,
  parameter m = 13,
  parameter t = 119,
  parameter L = 1547,

  // local:
  parameter K_elim = N*((K+N-1)/N),
  parameter mem_width = 1 << (m-`CLOG2(t+1)-1)
)
(
  input wire 							 clk,
  input wire 							 rst,
  input wire 							 pk_gen_start, // public key generation only starts after permutation and g(x) evaluations are all done
  input wire 							 eva_start,
//
// operations:
//  2'b00 = generate all entries
//  2'b01 = generate first part  (columns 0 to n-k)
//  2'b10 = generate second patr (columns n-k to n)
  input wire [1:0] 						 op,
//
  input wire [m*(t+1)-1:0] 				 eva_poly_in,
  output wire 							 mem_wr_en,
  output wire [`CLOG2(L*K_elim/N)-1 : 0] mem_wr_addr,
  output wire [N*M_elim-1:0] 			 mem_data_in,
  output wire 							 perm_rd_en,
  output wire [m-1:0] 					 perm_rd_addr,
  input wire [m-1:0] 					 perm_dout,
  output wire 							 init_done,
  output wire 							 eva_done,
`ifdef SHARED
  //ports for SHARED ADDITIVE FFT module
  output wire 							 fft_eva_start_kg,
  output wire [m*(t+1)-1:0] 			 fft_coeff_in_kg,
  output wire 							 fft_rd_en_kg,
  output wire [`CLOG2(t+1)-1:0] 		 fft_rd_addr_kg,
  input wire [2*m*mem_width-1:0] 		 fft_eva_dout_kg,
  input wire 							 fft_eva_done_kg,
`endif
 
  // added for testing purpose
  output wire [2*m*mem_width-1:0] 		 eva_dout
);

wire inv_start;
wire eva_rd_en;
wire [`CLOG2(t+1)-1:0] eva_rd_addr;
wire [N*m-1:0] index_in;
wire [N*m-1:0] eva_inv_in;
wire inv_done;
wire block_init_done;

wire [`CLOG2(K_elim/N)-1:0] block;
wire [`CLOG2(K_elim/N)-1:0] block_start;
wire [`CLOG2(K_elim/N)-1:0] block_end;
reg  [`CLOG2(K_elim/N)-1:0] block_buffer = 0;

assign block_start = (op == 2'b10) ? (m * t) / N : 0;
assign block_end   = (op == 2'b01) ? (m * t) / N - 1: K_elim/N - 1;

reg inv_start_buffer = 1'b0;
reg init_done_buffer = 0;

always @(posedge clk)
  begin
    inv_start_buffer <= (pk_gen_start | (block_init_done && (block < block_end)));

    block_buffer <= pk_gen_start ? block_start :
                    (block_init_done && (block <= block_end)) ? block_buffer + 1 : block_buffer;

    init_done_buffer <= (eva_start | pk_gen_start) ? 1'b0 : (block_init_done && (block == block_end));
  end

assign inv_start = inv_start_buffer;
assign block = block_buffer;
assign init_done = init_done_buffer;

inv_g_alpha #(.N(N), .K(K), .m(m), .t(t)) inv_g_alpha_inst (
  .clk(clk),
  .start(inv_start),
  .perm_dout(perm_dout),
  .eva_dout(eva_dout),
  .block(block),
  .perm_rd_en(perm_rd_en),
  .perm_rd_addr(perm_rd_addr),
  .eva_rd_en(eva_rd_en),
  .eva_rd_addr(eva_rd_addr),
  .index_out(index_in),
  .inv_out(eva_inv_in),
  .done(inv_done)
);

gen_H_entries #(.N(N), .L(L), .K(K_elim), .m(m)) gen_H_entries_inst (
  .clk(clk),
  .start(inv_done),
  .block(block),
  .index(index_in),
  .eva_inv(eva_inv_in),
  .wr_en(mem_wr_en),
  .data_in(mem_data_in),
  .wr_addr(mem_wr_addr),
  .block_init_done(block_init_done) 
);

`ifdef SHARED
assign fft_eva_start_kg  = eva_start;
assign fft_coeff_in_kg = eva_poly_in;
assign fft_rd_en_kg = eva_rd_en;
assign fft_rd_addr_kg = eva_rd_addr;
assign eva_dout = fft_eva_dout_kg;
assign eva_done = fft_eva_done_kg;
`else
add_FFT #(.numc(t+1), .gf(m), .mem_width(mem_width), .factor(0)) add_FFT_inst (
  .clk(clk),
  .start(eva_start),
  .coeff_in(eva_poly_in),
  .rd_en(eva_rd_en),
  .rd_addr(eva_rd_addr),
  .data_out(eva_dout),
  .done(eva_done) 
);
`endif
endmodule

