/*
 * function: generate module gen_H_entries
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

module gen_H_entries
#(
  parameter N = 20,         // N is the size of the column block
  parameter m = 13,         // 2^m is the size of the finite field, m=13
  parameter L = 100,        // L is the row number for GS
  parameter K = 260,         // K is the column number for GS
  // local: 
  //mul_tot is the total numer of multiplications perfromed per iteration
  parameter mul_tot = m*((N+m-1)/m)
)
(
  input wire clk,
  input wire start,
  input wire [`CLOG2(K/N)-1:0] block,
  input wire [N*m-1:0] index,
  input wire [N*m-1:0] eva_inv,
  output wire wr_en,
  output wire [N-1:0] data_in,
  output wire [`CLOG2(L*K/N)-1:0] wr_addr,
  output wire block_init_done
);
  
reg  [mul_tot*m-1:0] vec_a = 0;
reg  [mul_tot*m-1:0] vec_b = 0;
wire [mul_tot*m-1:0] vec_c;

// basis = {1, a, a^2, ..., a^12} instead of {a^12, a^11, ..., a, 1}
wire [N*m-1:0] alpha;
genvar j;
generate
  for (j=0; j < N*m; j=j+1)
    begin : gen_alpha
      assign alpha[j] = index[m*(j/m)+m-(j%m)-1];
    end
endgenerate
  
// Requires that vec_mul_multi takes m cycles!
reg  mul_start = 0;
wire mul_done;
reg  running = 0;

always @(posedge clk)
  begin
    vec_a <= start    ? {{(mul_tot-N)*m {1'b0}}, eva_inv} :
             mul_done ? vec_c :
             vec_a;

    vec_b <= start    ? {{(mul_tot-N)*m {1'b0}}, alpha} :
             vec_b;
  end

 
always @(posedge clk)
  begin
    running   <= start ? 1 :
                 block_init_done ? 0 :
                 running;

    mul_start <= (start | (mul_done && running));
  end

reg [mul_tot*m-1:0] dout_shift = 0;
reg [`CLOG2(m):0] counter = 0;
reg [`CLOG2(L*K/N)-1:0] wr_addr_buffer = 0;

always @(posedge clk)
  begin
    dout_shift     <= start ? {{(mul_tot-N)*m {1'b0}}, eva_inv} :
                      mul_done ? vec_c :
                      wr_en ? (dout_shift >> 1) : dout_shift; 

    counter        <= (start | mul_done) ? 0 :
                      (counter <= (m-1)) ? counter + 1 :
                      counter;

    wr_addr_buffer <= start ? L*block :
                      wr_en ? wr_addr_buffer + 1 : 
                      block_init_done ? 0 :
                      wr_addr_buffer;
  end

assign wr_en = running && (counter <= (m-1));

assign wr_addr = wr_addr_buffer;

assign block_init_done = (wr_addr == (L*(block+1)));
  
genvar i;
generate
  for (i=0; i < N; i=i+1)
    begin : gen_transpose
      assign data_in[i] = dout_shift[i*m];
    end
endgenerate
 
vec_mul_multi_H_gen vec_mul_multi_H_gen_inst (
  .clk(clk),
  .start(mul_start),
  .vec_a(vec_a),
  .vec_b(vec_b),
  .vec_c(vec_c),
  .done(mul_done)
);

endmodule

