/*
 * function: inverse the evalutation results.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/                     
 

module inv_g_alpha
#(
  parameter N = 20,     // the size of the column block
  parameter m = 13,     // the size of the finite field, m=13
  parameter K = 250,    // the column number
  parameter t = 119,    // number of correctable errors
  // local:
  parameter K_elim = N*((K+N-1)/N),
  parameter eva_mem_WIDTH = m*(1 << (m-`CLOG2(t+1))), //memory width in the add_FFT module
  parameter eva_mem_WIDTH_bits = m-`CLOG2(t+1),
  parameter eva_mem_DEPTH = (1 << `CLOG2(t+1)), //memory depth in the add_FFT module
  parameter eva_mem_DEPTH_bits = `CLOG2(t+1)
)
(
  input wire                         clk,
  input wire                         start,
  input wire [eva_mem_WIDTH-1:0]     eva_dout,
  input wire [`CLOG2(K_elim/N)-1:0]  block,
  output wire                        perm_rd_en,
  output wire [m-1:0]                perm_rd_addr, 
  input wire [m-1:0]                 perm_dout,
  output wire                        eva_rd_en,
  output wire [`CLOG2(t+1)-1:0]      eva_rd_addr,
  output wire [N*m-1:0]              index_out,
  output wire [N*m-1:0]              inv_out,
  output wire                        done 
);
 
reg [`CLOG2(N):0] counter = {(`CLOG2(N)+1){1'b0}};
  
always @(posedge clk)
  begin
    counter <= start ? {(`CLOG2(N)+1){1'b0}} :
               (counter < N) ? counter + 1 : 
               counter;
  end

reg running = 1'b0;
  
always @(posedge clk)
  begin
    running <= start ? 1'b1 :
               (counter >= (N-1)) ? 1'b0 : 
               running; 
  end
  
assign perm_rd_en = running;

reg [m-1:0] perm_rd_addr_buffer = {m{1'b0}};
  
always @(posedge clk)
  begin
    perm_rd_addr_buffer <= start ? N*block :
                            perm_rd_en ? perm_rd_addr_buffer + 1 : 
                            perm_rd_addr_buffer;
  end
  
assign perm_rd_addr = perm_rd_addr_buffer;

reg running_buffer = 1'b0;
reg eva_rd_en_buffer = 1'b0;
reg point_loc_start = 1'b0;
  
always @(posedge clk)
  begin
    running_buffer <= running;
    eva_rd_en_buffer <= running_buffer;
    point_loc_start <= eva_rd_en_buffer;
  end
  
assign eva_rd_en = eva_rd_en_buffer;

reg [eva_mem_DEPTH_bits-1:0] eva_rd_addr_buffer = {eva_mem_DEPTH_bits{1'b0}};
reg [eva_mem_WIDTH_bits-1:0] eva_rd_loc_buffer = {eva_mem_WIDTH_bits{1'b0}};
reg [eva_mem_WIDTH_bits-1:0] eva_rd_loc = {eva_mem_WIDTH_bits{1'b0}};
  
always @(posedge clk)
  begin
   eva_rd_addr_buffer <= (perm_dout >> eva_mem_WIDTH_bits);
   eva_rd_loc_buffer <= (perm_dout % (1 << eva_mem_WIDTH_bits));
   eva_rd_loc <= eva_rd_loc_buffer;
  end
  
assign eva_rd_addr = eva_rd_addr_buffer;

wire inv_done;
   
reg inv_done_buffer = 1'b0;
reg done_buffer = 1'b0;
reg done_buffer_2 = 1'b0;
  
always @(posedge clk)
  begin
    inv_done_buffer <= inv_done;
    
    done_buffer <= (!inv_done) && inv_done_buffer;
    done_buffer_2 <= done_buffer;
  end

assign done = done_buffer_2;

reg last_block = 0;
  
always @(posedge clk)
  begin
    last_block <= start ? (block == ((K_elim/N)-1)) : last_block;
  end

wire [m-1:0] inv_dout;

reg [N*m-1:0] inv_out_buffer = {N*m{1'b0}}; 
reg [N*m-1:0] inv_out_buffer_2 = {N*m{1'b0}}; 
  
always @(posedge clk)
  begin
    inv_out_buffer <= start ? {N*m{1'b0}} :
                      inv_done_buffer ? {inv_dout, inv_out_buffer[N*m-1:m]} : inv_out_buffer;
    
    inv_out_buffer_2 <= (!last_block) ? inv_out_buffer : 
                         {{(K_elim-K)*m {1'b0}}, inv_out_buffer[(N-K_elim+K)*m-1:0]};
  end

assign inv_out = inv_out_buffer_2;

reg [N*m-1:0] index_out_buffer = {N*m{1'b0}};
reg [N*m-1:0] index_out_buffer_2 = {N*m{1'b0}};
  
always @(posedge clk)
  begin
    index_out_buffer <= start ? 0 :
                        running_buffer ? {perm_dout, index_out_buffer[N*m-1:m]} : index_out_buffer;
    
    index_out_buffer_2 <= (!last_block) ? index_out_buffer : 
                          {{(K_elim-K)*m {1'b0}}, index_out_buffer[(N+K-K_elim)*m-1:0]};
  end

assign index_out = index_out_buffer_2;
 
wire inv_dout_en;

wire [m-1:0] inv_din;  
  
gf_inv #(.DELAY(1)) GFE_inv_H_gen_inst (
  .clk(clk),
  .din_A(inv_din),
  .dout_A(inv_dout),
  .dout_en_A(inv_dout_en),
  .din_B(0),
  .dout_B(),
  .dout_en_B()
);
   
point_loc point_loc_inst (
  .start(point_loc_start),
  .clk(clk),
  .location(eva_rd_loc),
  .data_in(eva_dout),
  .data_out(inv_din),
  .done(inv_done)
);

endmodule

