/*
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

// public key generation module
// input: goppa polynomial
//        permutation memory interface
// output: public key memory interface

module gen_pk
#(
  parameter N = 40,
  parameter L = 1547,
  parameter K = 8192,
  parameter M_elim = 1,
  parameter m = 13,
  parameter t = 119,
  parameter BLOCK = 20,
  parameter DATA = "", 
  // local:
  parameter K_elim = N*((K+N-1)/N),
  parameter mem_width = (1 << (m-`CLOG2(t+1)-1)) // actual memory width in add_FFT, width = 2*mem_width*m
)
(
  input wire 						  clk,
  input wire 						  rst,
  input wire 						  pk_gen_start,
  input wire 						  eva_start, // start signal
  input wire [m*(t+1)-1:0] 			  eva_poly_in,
 
  // outside reads have priorities
  input wire 						  rd_en,
  input wire [`CLOG2(L*K_elim/N)-1:0] rd_addr,
 
  output wire 						  elim_done,
  output wire [N*M_elim-1:0] 		  elim_data_out,
  output wire 						  fail,
  
  output wire 						  perm_rd_en,
  output wire [m-1:0] 				  perm_rd_addr,
  input wire [m-1:0] 				  perm_dout,
  
  // added for testing purpose
  output wire [2*m*mem_width-1:0] 	  eva_dout,
  output wire 						  init_done,
`ifdef SHARED
    //ports for SHARED ADDITIVE FFT module
  output wire 						  fft_eva_start_kg,
  output wire [m*(t+1)-1:0] 		  fft_coeff_in_kg,
  output wire 						  fft_rd_en_kg,
  output wire [`CLOG2(t+1)-1:0] 	  fft_rd_addr_kg,
  input wire [2*m*mem_width-1:0] 	  fft_eva_dout_kg,
  input wire 						  fft_eva_done_kg,

  input wire 						  host_pk_wr_en,
  input wire [`CLOG2(L*K_elim/N)-1:0] host_pk_wr_addr,
  input wire [N*M_elim-1:0] 		  pk_din,
`endif
  output wire 						  eva_done
);

wire mem_wr_en;
wire [`CLOG2(L*K_elim/N)-1:0] mem_wr_addr;
wire [N*M_elim-1:0] mem_data_in;

reg wr_en_buffer = 1'b0;
reg [`CLOG2(L*K_elim/N)-1:0] wr_addr_buffer = 1'b0;
reg [N*M_elim-1:0] data_in_buffer = 1'b0;
reg GS_start = 1'b0;

reg init_done_running_buffer = 1'b0;

wire success;
reg start_right = 1'b0;

wire [1:0] gen_left_op;
wire [1:0] gen_right_op;

wire [1:0] gen_op;
reg [1:0] gen_op_reg = 0;

assign gen_op = pk_gen_start ? gen_left_op : 
                 success ? gen_right_op : 
                 gen_op_reg;

always @(posedge clk)
  begin
     gen_op_reg <= gen_op;
  end


gen_H #(.N(N), .L(L), .K(K), .M_elim(M_elim), .m(m), .t(t)) gen_H_inst (
  .clk(clk),
  .rst(rst),
  .op(gen_op),
  .pk_gen_start(pk_gen_start | success),
  .eva_start(eva_start),
  .eva_poly_in(eva_poly_in),
  .mem_wr_en(mem_wr_en),
  .mem_wr_addr(mem_wr_addr),
  .mem_data_in(mem_data_in),
  .perm_rd_en(perm_rd_en),
  .perm_rd_addr(perm_rd_addr),
  .perm_dout(perm_dout),
  .init_done(init_done),
  .eva_done(eva_done),
`ifdef SHARED
  //shared fft signals
  .fft_eva_start_kg(fft_eva_start_kg),
  .fft_coeff_in_kg(fft_coeff_in_kg),
  .fft_rd_en_kg(fft_rd_en_kg),
  .fft_rd_addr_kg(fft_rd_addr_kg),
  .fft_eva_dout_kg(fft_eva_dout_kg),
  .fft_eva_done_kg(fft_eva_done_kg),
`endif
  .eva_dout(eva_dout)
);
 
wire init_done_running = init_done | init_done_running_buffer;


reg elim_right_running = 1'b0;
 
always @(posedge clk)
  begin
    init_done_running_buffer <= eva_start ? 1'b0 :
                                (success | fail | elim_done) ? 1'b0 : 
                                init_done ? 1'b1 : init_done_running_buffer;

    elim_right_running <= pk_gen_start ? 1'b0 : 
                          success ? 1'b1 : 
                          elim_done ? 1'b0 :
                          elim_right_running;
  end

always @(posedge clk)
  begin
    GS_start <= init_done && !elim_right_running;
    start_right <= init_done && elim_right_running;

    wr_en_buffer <= ((!init_done_running) && mem_wr_en);
    wr_addr_buffer <= mem_wr_addr;
    data_in_buffer <= mem_data_in;
  end

    
systemizer #(.N(N), .L(L), .K(K_elim), .M(M_elim), .BLOCK(BLOCK), .DATA(DATA)) systemizer_inst (
  .clk(clk),
  .rst(rst),
  .gen_left_op(gen_left_op),
  .gen_right_op(gen_right_op),
  .start(GS_start),
  .fail(fail),
  .success(success),
  .start_right(start_right),
  .done(elim_done),
  .rd_en(rd_en),
  .rd_addr(rd_addr),
  .data_out(elim_data_out),
  .wr_en(wr_en_buffer),
  .wr_addr(wr_addr_buffer),
`ifdef SHARED
  // public key input used in joint design 
  .host_pk_wr_en(host_pk_wr_en),
  .host_pk_wr_addr(host_pk_wr_addr),
  .pk_din(pk_din),
`endif
  .data_in(data_in_buffer)
);

// elim_H_elim #(.S(N), .MT(L), .N(K_elim), .DATA("")) elim_inst (
//   .clk(clk),
//   .rst(rst),
//   .elim_start(GS_start),
//   .elim_done(elim_done),
//   .fail(fail),
//   .rd_en(rd_en),
//   .rd_addr(rd_addr),
//   .data_out(elim_data_out),
//   .wr_en(wr_en_buffer),
//   .wr_addr(wr_addr_buffer),
//   .data_in(data_in_buffer)
// );

endmodule

