// secret polynomial generation module
// input: random element r in (GF(2^m))^t
// output: polynomial memory interface
/*
 * This file generates the secret irreducible polynomial.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


module gen_poly
#(
  parameter N = 2,
  parameter M = 13,
  parameter L = 119,
  parameter K = 120,
  parameter BLOCK = 1,
  // local:
  parameter K_elim = N*((K+N-1)/N) // in Gaussian systemizer, need to make sure K is a multiple of N
)
(
  input wire clk,
  input wire start,
  input wire rst,
  input wire [(L*M-1):0] coeff_in,
  
  // read requests have priorities
  input wire rd_en,
  input wire [`CLOG2(L*K_elim/N) - 1 : 0] rd_addr,
  
  output wire [(N*M)-1 : 0] data_out,
  output wire done,
  output wire init_done
);

wire wr_en;
wire [`CLOG2(L*K_elim/N) - 1 : 0] wr_addr;
wire [(N*M)-1:0] data_in;

reg GS_start = 0;

reg wr_en_buffer = 0;
reg [`CLOG2(L*K_elim/N) - 1 : 0] wr_addr_buffer = 0;
reg [(N*M)-1:0] data_in_buffer = 0;

always @(posedge clk)
  begin
    wr_en_buffer <= wr_en;
	wr_addr_buffer <= wr_addr;
	data_in_buffer <= data_in;
  end

always @(posedge clk)
  begin
    GS_start <= init_done;
  end
  
 
mem_init_R #(.N(N), .L(L), .K(K), .M(M)) mem_init_R_inst (
  .clk(clk),
  .start(start),
  .coeff_in(coeff_in),
  .wr_en(wr_en),
  .wr_addr(wr_addr),
  .data_in(data_in),
  .init_done(init_done) 
);
 
elim_R_elim #(.N(N), .L(L), .K(K_elim), .M(M), .BLOCK(BLOCK), .DATA("")) elim_inst (
  .clk(clk),
  .rst(rst),
  .start(GS_start), //changed to check the initialization value of R
  .start_right(1'b0),
  .done(done),
  .rd_en(rd_en),
  .rd_addr(rd_addr),
  .data_out(data_out),
  .wr_en(wr_en_buffer),
  .wr_addr(wr_addr_buffer),
  .data_in(data_in_buffer)
);
 
endmodule
