/*
 * This file computes the matrix for minimal polynomial computation.
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


module mem_init_R
#(
  parameter N = 4,
  parameter M = 13,
  parameter L = 119, // number of correctable errors
  parameter K = 120,
  // local:
  parameter K_elim = N*((K+N-1)/N)
)
(
  input   wire                            clk,
  input   wire                            start,
  input   wire [(L*M-1):0]                coeff_in,
  output  wire                            wr_en,
  output  wire [`CLOG2(L*K_elim/N)-1 : 0] wr_addr,
  output  wire [(N*M)-1 : 0]              data_in,
  output  wire                            init_done
);

// inputs for the polynomial multiplier
reg [(L*M)-1:0] poly_g_in [1:N];
reg [(L*M)-1:0] poly_f_in [1:N];
wire [(L*M)-1:0] poly_out [1:N];
  
// stores the results from the first phase: (1, r, r^2, .., r^(N-1))
reg [(L*M)-1:0] phase_1_out [1:N];  
reg [(L*M)-1:0] poly_out_shift [1:N];

//initialize poly_g_in, poly_f_in and phase_1_out.
integer j;
initial 
  for (j = 1; j <= N; j = j + 1)
    begin
      poly_g_in[j] = {L*M{1'b0}};
      poly_f_in[j] = {L*M{1'b0}};
      phase_1_out[j] = {L*M{1'b0}};
      poly_out_shift[j] = {L*M{1'b0}};
    end

// control signals of the polynomial multiplier
reg mul_start = 1'b0;
wire mul_done_out [1:N];
wire mul_done;
assign mul_done = mul_done_out[1];

reg running = 1'b0;

always @(posedge clk)
  begin
    running <= start ? 1'b1 :
               init_done ? 1'b0 :
               running;
    
    mul_start <= (start | (running && mul_done));
  end

  
// counter for the 1st phase
reg [`CLOG2(N)+1:0] phase_1_counter = 0;

reg phase_2 = 1'b0;

wire phase_1_done;
assign phase_1_done = ((phase_1_counter == N) && mul_done);

always @(posedge clk)
  begin
    phase_1_counter <= start ? 1 :
                       (mul_done && (phase_1_counter <= N)) ? phase_1_counter + 1 :  
                        phase_1_counter;                                     
  end

// buffer results during the 1st phase
// the first element is always 1
     
always @(posedge clk)
  begin
    phase_1_out[1] <= start ? 1 : 
                      phase_1_out[1]; // only the first poly_mul is working
  end

genvar t;
generate 
  for (t = 2; t <= N; t = t + 1)
    begin: gen_phase_1_out
      always @(posedge clk)
        begin
          phase_1_out[t] <= start ? 0 :
                            ((phase_1_counter == (t-1)) && mul_done) ? poly_out[1] :
                            phase_1_out[t];
        end
    end
endgenerate

// the 1st phase is running
reg phase_1 = 0;
always @(posedge clk)
  begin
    phase_1 <= start ? 1'b1 :
                phase_1_done ? 1'b0 :
                //init_done ? 1'b0 :
                phase_1;
  end

// poly_g_in 
genvar q;
generate
  for (q = 1; q <= N; q = q + 1)
    begin: gen_poly_g_in
      always @(posedge clk)
        begin
          poly_g_in[q] <= start ? 1 : // the first mul in phase 1
                          phase_1_done ? phase_1_out[q] : // the first mul in phase 2 
                          mul_done ? poly_out[q] :
                          poly_g_in[q]; 
        end
    end
endgenerate

// poly_f_in
genvar p;
generate
  for (p = 1; p <= N; p = p+1)
    begin: gen_poly_f_in
      always @(posedge clk)
        poly_f_in[p] <= start ? coeff_in :
                        phase_1_done ? poly_out[1] : // at this time, poly_out[1] = r^N
                         poly_f_in[p];
    end
endgenerate 

// write to memory
genvar g;
generate 
  for (g = 1; g <= N; g = g+1)
    begin: gen_poly_out_shift
      always @(posedge clk)
        begin
          poly_out_shift[g] <= phase_1_done ? phase_1_out[g] :
                              (phase_2 && mul_done) ? poly_out[g] :
                              (poly_out_shift[g] >> M);
        end
    end
endgenerate
  
   wire [(L*M)-1:0] poly_out_shift_test;
  assign poly_out_shift_test = poly_out_shift[1];
  wire [(L*M)-1:0] poly_g_in_test;
  assign poly_g_in_test = poly_g_in[1];
  wire [(L*M)-1:0] poly_f_in_test;
  assign poly_f_in_test = poly_f_in[1];
  wire [(L*M)-1:0] poly_out_test;
  assign poly_out_test = poly_out[1];

genvar a;
generate
  for (a = 0; a < N; a = a+1)
    begin: gen_data_in
      assign data_in[a*M+M-1:a*M] = poly_out_shift[a+1][M-1:0];
    end
endgenerate

always @(posedge clk)
  begin
    phase_2 <= start ? 1'b0 :
              phase_1_done ? 1'b1 :
              init_done ? 1'b0 :
              phase_2;
  end
  
reg [`CLOG2(L):0] wr_to_mem_counter = 0; 

always @(posedge clk)
  begin
    wr_to_mem_counter <= (start | mul_done) ? 0 : 
                         (phase_2 && (wr_to_mem_counter < L)) ? wr_to_mem_counter + 1 :
                         init_done ? 0 :
                         wr_to_mem_counter;
  end

assign wr_en = phase_2 && (wr_addr < L*K_elim/N) && (wr_to_mem_counter < L);  

reg [`CLOG2(L*K_elim/N) - 1 : 0] wr_addr_buffer = 0;	

always @(posedge clk)
  begin
    wr_addr_buffer <= start ? 0 :
					             wr_en ? wr_addr_buffer + 1 : wr_addr_buffer;
  end 

assign wr_addr = wr_addr_buffer;

// start signal for elim module
reg init_done_buffer = 0; 
always @(posedge clk) 
  begin
    init_done_buffer <= start ? 0 : (wr_addr == (L*K_elim/N-1));
  end

assign init_done = init_done_buffer;
  
genvar i;
generate
  for (i=1; i <= N; i=i+1) 
    begin: gen_poly_mul_reduced_instances
      poly_mul_reduced poly_mul_reduced_inst(
        .start(mul_start),
        .clk(clk),
        .poly_g(poly_g_in[i]),
        .poly_f(poly_f_in[i]),
        .poly_out(poly_out[i]),
        .done(mul_done_out[i])
      );
    end
endgenerate

endmodule
