/*
 *  
 *
 * Copyright (C) 2018
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

module key_gen 
  #(
    // security parameters
    parameter M = 13, // 2^m is the size of the field
    parameter T = 119, // number of correctable erros
    parameter K = 6960, // code length
    
    // system parameters
    // key generation
    parameter N_H = 20, // N in H elimination
    parameter BLOCK_H = 10, // BLOCK in H elimination
    parameter N_R = 2, // N in R elimination
    parameter BLOCK_R = 1, // BLOCK in R elimination

    parameter SIGMA1 = 32, // bit width of random values for generation the polynomial
    parameter SIGMA2 = 32, // bit width of integer used in sort module

    parameter MEM_WIDTH = 32,
   
    parameter L = M*T,

	parameter DATA = "",
`ifdef SHARED
    //paramter for SHARED FFT module
    parameter FFT_mem_width = 1 << (M-`CLOG2(T+1)-1),
`endif
    parameter K_H_elim = N_H*((K+N_H-1)/N_H),
    parameter K_R_elim = N_R*((T+1+N_R-1)/N_R)
  )
  (
    // general inputs
    input wire 								clk,
    input wire 								start, // general start signal
    input wire [1:0] 						request, 
    // Simple instructions
    // 00: given two seeds, return a key pair
    
    // polynomial generation
    input wire 								host_beta_rand_wr_en, 
    input wire [`CLOG2(T)-1:0] 				host_beta_rand_wr_addr,
    input wire [SIGMA1-1:0] 				host_beta_rand_data_in,
    //
    input wire 								host_sk_poly_rd_en,
    input wire [`CLOG2(T*K_R_elim/N_R)-1:0] host_sk_poly_rd_addr,
    output wire [M*N_R-1:0] 				sk_poly_dout,
    output wire 							sk_poly_gen_done,

`ifndef SHARED
    // permutation generation
    input wire 								host_alpha_rand_wr_en, 
    input wire [M-1:0] 						host_alpha_rand_wr_addr,
    input wire [SIGMA2-1:0] 				host_alpha_rand_data_in,
`endif
    input wire 								host_perm_rd_en,
    input wire [M-1:0] 						host_perm_rd_addr,
`ifndef SHARED
    output wire [M-1:0] 					perm_dout,
    output wire 							perm_done,
`endif
`ifdef SHARED
    // shared fieldordering generation  (secret permutation generator)
    output wire 							sk_gen_start,
    input wire 								perm_done,
    input wire 								field_ordering_fail,
    output wire 							perm_rd_en_valid,
    output wire [M-1:0] 					perm_rd_addr_valid,
    input wire [M-1:0] 						perm_dout,
    input wire [SIGMA2-1:0] 				perm_rand_out,

    //ports for SHARED ADDITIVE FFT module
    output wire 							fft_eva_start_kg,
    output wire [M*(T+1)-1:0] 				fft_coeff_in_kg,
    output wire 							fft_rd_en_kg,
    output wire [`CLOG2(T+1)-1:0] 			fft_rd_addr_kg,
    input wire [2*M*FFT_mem_width-1:0] 		fft_eva_dout_kg,
    input wire 								fft_eva_done_kg,

    input wire 								host_pk_wr_en,
    input wire [`CLOG2(L*K_H_elim/N_H)-1:0] host_pk_wr_addr, 
    input wire [N_H-1:0] 					host_pk_din,
`endif
    // PK generation
    input wire 								host_pk_rd_en,
    input wire [`CLOG2(L*K_H_elim/N_H)-1:0] host_pk_rd_addr, 
    output wire [N_H-1:0] 					pk_dout,
    output wire 							pk_gen_done,
    output wire 							pk_gen_fail
  );

`ifndef SHARED
  wire sk_gen_start;
`endif

  assign sk_gen_start = (start && (request == 2'b00));
  
  wire R_elim_init_done;
  wire sk_poly_rd_en_valid;
  wire [`CLOG2(T*K_R_elim/N_R)-1:0] sk_poly_rd_addr_valid;
  
  reg pk_sk_poly_rd_en = 1'b0;
  reg pk_sk_poly_rd_en_buf = 1'b0; 
  reg [`CLOG2(T*K_R_elim/N_R)-1:0] pk_sk_poly_rd_addr = {(`CLOG2(T*K_R_elim/N_R)){1'b0}};
  reg [`CLOG2(T):0] pk_sk_poly_rd_counter = {(`CLOG2(T)+1){1'b0}};
  reg buf_sk_poly_done_pre = 1'b0;
  reg buf_sk_poly_done = 1'b0;
  reg [M*T-1:0] buf_sk_poly = {M*T{1'b0}};

  
  reg [M-1:0] r_gen_coeff_mem [0:T-1];
  wire [M*T-1:0] r_gen_coeff_out;

  // access r in full width
  genvar t;
  generate
    for (t = 0; t < T; t = t + 1)
      assign r_gen_coeff_out[M*(t+1)-1:M*t] = r_gen_coeff_mem[T-1-t];
  endgenerate

  always @ (posedge clk)
  begin
    if (host_beta_rand_wr_en)
      r_gen_coeff_mem[host_beta_rand_wr_addr] <= host_beta_rand_data_in[M-1:0];
  end


  always @(posedge clk)
    begin
      
      pk_sk_poly_rd_counter <= sk_gen_start ? {(`CLOG2(T)+1){1'b0}} :
                               (pk_sk_poly_rd_en && (pk_sk_poly_rd_counter < T)) ? pk_sk_poly_rd_counter + 1 :
                               pk_sk_poly_rd_counter;
      
      pk_sk_poly_rd_en <= (sk_gen_start | (pk_sk_poly_rd_counter >= (T-1))) ? 1'b0 :
                          sk_poly_gen_done ? 1'b1 :
                          pk_sk_poly_rd_en;
      
      pk_sk_poly_rd_en_buf <= pk_sk_poly_rd_en;
      
      pk_sk_poly_rd_addr <= sk_poly_gen_done ? (T*K_R_elim/N_R-1) :
                            pk_sk_poly_rd_en ? pk_sk_poly_rd_addr - 1 :
                            pk_sk_poly_rd_addr;
      
      buf_sk_poly <= sk_poly_gen_done ? {M*T{1'b0}} :
                      pk_sk_poly_rd_en_buf ? {buf_sk_poly[M*(T-1)-1:0], sk_poly_dout[M*(N_R-(K_R_elim-(T+1)))-1:M*(N_R-(K_R_elim-(T+1))-1)]} :
                      buf_sk_poly;
      
      buf_sk_poly_done_pre <= (pk_sk_poly_rd_counter == (T-1));
      buf_sk_poly_done <= buf_sk_poly_done_pre;
                            
    end
  
  wire eva_done;
  reg eva_done_stay = 1'b0;
  reg perm_done_stay = 1'b0;
  wire [2*M*MEM_WIDTH-1:0] eva_dout;
  reg pk_gen_start = 1'b0;
  reg pk_gen_running = 1'b0;
  reg pk_gen_done_stay = 1'b0;
  
  always @(posedge clk)
    begin
      pk_gen_running <= (sk_gen_start | pk_gen_done) ? 1'b0 :
                        sk_poly_gen_done ? 1'b1 :
                        pk_gen_running;
      
      pk_gen_done_stay <= sk_gen_start ? 1'b0 :
                          pk_gen_done ? 1'b1 :
                          pk_gen_done_stay;
      
      eva_done_stay <= (sk_gen_start | pk_gen_done) ? 1'b0 :
                        eva_done ? 1'b1 :
                        eva_done_stay;
      
      perm_done_stay <= (sk_gen_start | pk_gen_done) ? 1'b0 :
                        perm_done ? 1'b1 :
                        perm_done_stay;
      
      pk_gen_start <= ((eva_done_stay && perm_done) || (eva_done) && perm_done_stay);
      
    end
  
  assign sk_poly_rd_en_valid = pk_gen_running ? pk_sk_poly_rd_en :  
                               pk_gen_done_stay ? host_sk_poly_rd_en :
                               1'b0;
  
  assign sk_poly_rd_addr_valid = pk_gen_running ? pk_sk_poly_rd_addr :
                                 pk_gen_done_stay ? host_sk_poly_rd_addr :
                                 {(`CLOG2(T*K_R_elim/N_R)){1'b0}};
  
  wire sort_init_done;
`ifndef SHARED
  wire [SIGMA2-1:0] perm_rand_out;
  wire perm_rd_en_valid;
  wire [M-1:0] perm_rd_addr_valid;
`endif
  
  wire pk_perm_rd_en;
  wire [M-1:0] pk_perm_rd_addr;
  
  assign perm_rd_en_valid = pk_gen_running ? pk_perm_rd_en :
                            pk_gen_done_stay ? host_perm_rd_en :
                            1'b0;
  
  assign perm_rd_addr_valid = pk_gen_running ? pk_perm_rd_addr :
                              pk_gen_done_stay ? host_perm_rd_addr :
                              {M{1'b0}};
  
  wire H_elim_init_done;
  
  gen_poly #(.M(M), .L(T), .K(T+1), .N(N_R), .BLOCK(BLOCK_R)) poly_gen_inst (
    .clk(clk),
    .start(sk_gen_start),
    .rst(1'b0),
    .coeff_in(r_gen_coeff_out),
    .rd_en(sk_poly_rd_en_valid),
    .rd_addr(sk_poly_rd_addr_valid),
    .data_out(sk_poly_dout),
    .done(sk_poly_gen_done),
    .init_done(R_elim_init_done)
  );
  
  wire H_elim_fail;
`ifndef SHARED
  wire field_ordering_fail;
`endif
  assign pk_gen_fail = H_elim_fail | field_ordering_fail;

`ifndef SHARED
   // secret permutation generator
   field_ordering #(.M(M), .SIGMA2(SIGMA2)) perm_gen_inst (
    .clk(clk),
    .start(sk_gen_start),  
    .wr_en(host_alpha_rand_wr_en),  
    .wr_addr(host_alpha_rand_wr_addr),
    .rand_in(host_alpha_rand_data_in),
    .done(perm_done),
    .fail(field_ordering_fail),
    .rd_en(perm_rd_en_valid),
    .rd_addr(perm_rd_addr_valid),
    .index_out(perm_dout),
    .rand_dout(perm_rand_out)
  );
`endif

  reg [M-1:0] eva_coeff_in_one = 1;

 
  // public key generator
  gen_pk #(.N(N_H), .L(L), .K(K), .M_elim(1), .m(M), .t(T), .BLOCK(BLOCK_H), .DATA(DATA)) pk_gen_inst (
    .clk(clk),
    .rst(1'b0),
    .pk_gen_start(pk_gen_start),
    .eva_start(buf_sk_poly_done),
    .eva_poly_in({eva_coeff_in_one, buf_sk_poly}),
    .rd_en(host_pk_rd_en),
    .rd_addr(host_pk_rd_addr),
    .elim_done(pk_gen_done),
    .elim_data_out(pk_dout),
    .fail(H_elim_fail),
    .perm_rd_en(pk_perm_rd_en),
    .perm_rd_addr(pk_perm_rd_addr),
    .perm_dout(perm_dout),
    .eva_dout(eva_dout),
    .init_done(H_elim_init_done),
`ifdef SHARED
     //shared fft signals
    .fft_eva_start_kg(fft_eva_start_kg),
    .fft_coeff_in_kg(fft_coeff_in_kg),
    .fft_rd_en_kg(fft_rd_en_kg),
    .fft_rd_addr_kg(fft_rd_addr_kg),
    .fft_eva_dout_kg(fft_eva_dout_kg),
    .fft_eva_done_kg(fft_eva_done_kg),

	.host_pk_wr_en(host_pk_wr_en),
    .host_pk_wr_addr(host_pk_wr_addr),
    .pk_din(host_pk_din),
`endif
    .eva_done(eva_done)
  );
  
endmodule

