#
# Function: compute theoretical result for key_gen.
# Copyright (C) 2018
# Authors: Wen Wang <wen.wang.ww349@yale.edu>
#          Ruben Niederhagen <ruben@polycephaly.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import random
import argparse
import textwrap

parser = argparse.ArgumentParser(description='Generate memory contents for H.',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)

def auto_int(x):
    return int(x, 0)
 
parser.add_argument('-t', dest='t', type=int, default=119,  
          help='number of errors; default = 119')
parser.add_argument('-N', dest='N', type=int, default=20,  
          help='column block size') 
parser.add_argument('-n', dest='n', type=int, default=8192,  
          help='code length') 
parser.add_argument('-m', '--gf', dest='m', type=int, default=13,
          help='field GF(2^m); default = GF(2^13)')
parser.add_argument('-b', dest='b', type=int, default=32,
          help='bit width of random number used in sort module; default = 32')
parser.add_argument('-s1', '--sigma1', dest='sigma1', type=int, default=16,
          help='number of bits for generating the polynomial')
parser.add_argument('-s2', '--sigma2', dest='sigma2', type=int, default=32,
          help='number of bits from PRNG to be compared')
parser.add_argument('-s', '--seed', dest='seed', type=int,default=None,
          help='pyhton rng seed')
parser.add_argument('-d', '--delta', dest='delta', type=auto_int, default=None,
          help='initial delta seed')
parser.add_argument('--fail', dest='fail', action='store_true', default=False,
          help='fail in H systemize')
parser.add_argument('-o', '--output	', dest='fileout', default='key_sage.out',
          help='output file')
parser.add_argument('--memseed', dest='fileseed', default='mem_256.out',
          help='output file')
 
args = parser.parse_args()

if args.seed:
  random.seed(args.seed)


t = args.t
m = args.m
n = args.n
#b = args.b

F2 = GF(2)
F2z.<z> = F2[]
F2zy.<y> = F2z[]

f = {
   9 : z^9+z^1+1, # for testing
  10 : z^10+z^3+1, # for testing
  12 : z^12+z^3+1,
  13 : z^13+z^4+z^3+z+1,
}

F = {
   20 : z^20+z^3+1, # for testing
   50 : y^50+y^5+y^2+z, # for testing
   64 : y^64+y^3+y+z,
   96 : y^96+y^10+y^9+y^6+1,
  119 : y^119+y^8+1,
  128 : y^128+y^7+y^2+y+1,
}


# ----- derived parameters

q = 2^m
k = n-m*t
Fq.<zinFq> = GF(q,name='zinFq',modulus=f[m])
Fqt.<yinFqt> = Fq.extension(F[t])

# ----- symmetric-cryptography parameters

l = 256
sigma1 = 16
sigma2 = 32

mu,nu = 0,0

Fqx.<x> = Fq[]


import operator
def xor(a,b):
  return operator.xor(a,b)

def keccak(s):
  R = 1

  def complement(a):
    return xor(-1,a)
  
  def rol(a,b):
    return 0xffffffffffffffff & ((a<<b)|(a>>(64-b)))

  def rL(x,y):
    return sum(s[8*(x+5*y)+i]<<(8*i) for i in range(8))

  def wL(x,y,u):
    for i in range(8):
      s[8*(x+5*y)+i] = 255&(u>>(8*i))

  def XL(x,y,u):
    for i in range(8):
      s[8*(x+5*y)+i] = xor(s[8*(x+5*y)+i],255&(u>>(8*i)))

  for i in range(24):
    C = [None]*5
    for x in range(5):
      C[x] = xor(xor(xor(rL(x,0),rL(x,1)),xor(rL(x,2),rL(x,3))),rL(x,4))
    for x in range(5):
      D = xor(C[(x+4)%5],rol(C[(x+1)%5],1))
      for y in range(5):
        XL(x,y,D)
    x,y,r = 1,0,0
    D = rL(x,y)
    for j in range(24):
      r += j+1
      x,y = y,(2*x+3*y)%5
      C[0] = rL(x,y)
      wL(x,y,rol(D,r%64))
      D = C[0]

    for y in range(5):
      C = [None]*5
      for x in range(5):
        C[x] = rL(x,y)
      for x in range(5):
        wL(x,y,xor(C[x],complement(C[(x+1)%5])&C[(x+2)%5]))
    for j in range(7):
      R = xor(R<<1,0x71 if R&0x80 else 0)
      if (R&2)>>1:
        XL(0,0,1<<((1<<j)-1))

def shake256(input,outlen):
  s = [0]*200
  b = 0
  while len(input) > 0:
    for inputb in bytearray(input[:136]):
      s[b] = xor(s[b],inputb)
      b += 1
    if b == 136:
      keccak(s)
      b = 0
    input = input[136:]
  s[b] = xor(s[b],0x1f)
  s[136-1] = xor(s[136-1],0x80)

  result = b''
  while outlen > 0:
    keccak(s)
    result += bytes(bytearray(s[:136][:outlen]))
    outlen -= 136
  return result


def G(delta):
  assert len(delta) == 32
  x = b'@'+delta
  assert len(x) == 33
  return shake256(x,(n+sigma2*q+sigma1*t+l)//8)

# spec numbers rows and columns from 1
# this code numbers rows and columns from 0
def reduce(X):
  R = X.echelon_form()
  assert R.row_space() == X.row_space()
  r = R.rank()
  c = []
  for i in range(r):
    ok = False
    for j in range(R.ncols()):
      if R[i][j] != 0:
        ok = True
        c += [j]
        break
    assert ok
  assert r <= R.nrows()
  for i in range(r+1,R.nrows()):
    assert R[i].is_zero()
  assert all(c[i] < c[i+1] for i in range(r-1))
  for i in range(r):
    for j in range(R.nrows()):
      assert (R[j][c[i]] != 0) == (j == i)
  return R,c


def is_bit_vector(e,n):
  if len(e) != n: return False
  return all(parent(ei) == F2 for ei in e)

def bytes_from_vector(v):
  r = len(v)
  assert is_bit_vector(v,r)
  expectedlen = ceil(r/8)
  vpad = copy(v)
  while len(vpad)%8: vpad += [F2(0)]
  assert len(vpad) == 8*expectedlen
  result = [sum(int(vpad[8*i+j])<<j for j in range(8)) for i in range(expectedlen)]
  result = bytes(bytearray(result))
  assert len(result) == expectedlen
  return result

def vector_from_bytes(b,r):
  assert 8*len(b)-8 < r
  assert r <= 8*len(b)
  v = list(bytearray(b))
  v = [(v[i]>>j)&1 for i in range(len(b)) for j in range(8)]
  v = [F2(vj) for vj in v]
  assert len(v) == 8*len(b)
  v = v[:r]
  assert is_bit_vector(v,r)
  return v

def is_allowed_polynomial(g):
  if parent(g) != Fqx: return False
  if not g.is_monic(): return False
  if not g.is_irreducible(): return False
  return g.degree() == t

def is_fieldordering(alpha):
  if len(alpha) != q: return False
  if len(set(alpha)) != q: return False
  return all(parent(alphaj) == Fq for alphaj in alpha)

def is_support(alpha):
  if len(alpha) != n: return False
  if len(set(alpha)) != n: return False
  return all(parent(alphaj) == Fq for alphaj in alpha)

# R,c must be returned by reduce()
def is_semi_systematic(R,c,mu,nu):
  assert mu == int(mu)
  assert nu == int(nu)
  assert mu >= 0
  assert nu >= 0

  r = len(c)
  assert r >= mu
  assert R.ncols() >= r-mu+nu

  result1 = (R.nrows() == r)
  result2 = all(c[i] == i for i in range(r-mu))
  result3 = all(c[i] <= i-mu+nu for i in range(r))
  if r > mu:
    assert result2 == (c[r-mu-1] == r-mu-1)
    assert result3 == (c[r-1] <= r-mu+nu-1)

  result = result1 and result2 and result3
  if mu == nu: assert result == is_systematic(R,c)
  return result

def is_publickey(T):
  return parent(T) == MatrixSpace(F2,n-k,k)

# R,c must be returned by reduce()
def is_systematic(R,c):
  r = len(c)
  result = (c == list(range(r)))
  if r > 0: assert result == (c[r-1] == r-1)
  if r != R.nrows(): result = False
  assert result == (R.matrix_from_columns(range(r)) == identity_matrix(r))
  return result

def is_columnselection(c):
  return (len(c) == mu
      and len(set(c)) == mu
      and c == tuple(sorted(c))
      and all(parent(ci) == int for ci in c)
      and all(n-k-mu <= ci for ci in c)
      and all(ci < n-k-mu+nu for ci in c)
  )

def SystematicMatGen(Gamma):
  g,alpha = Gamma
  assert is_allowed_polynomial(g)
  assert is_support(alpha)

  recipgalpha = [1/g(alphaj) for alphaj in alpha]

  Hhat = matrix(F2,m*t,n)
  for i in range(t):
    for j in range(n):
      hij = alpha[j]^i*recipgalpha[j] # Htwiddle[i][j]
      hijbits = hij.polynomial()
      for pos in range(m):
        Hhat[i*m+pos,j] = hijbits[pos]

  H,c = reduce(Hhat)
  if not is_systematic(H,c): return False

  T = H.matrix_from_columns(range(n-k,n))
  assert is_publickey(T)
  return T,Gamma

def MatGen(Gamma):
  g,alpha = Gamma
  assert is_allowed_polynomial(g)
  assert is_support(alpha)

  recipgalpha = [1/g(alphaj) for alphaj in alpha]

  Hhat = matrix(F2,m*t,n)
  for i in range(t):
    for j in range(n):
      hij = alpha[j]^i*recipgalpha[j] # Htwiddle[i][j]
      hijbits = hij.polynomial()
      for pos in range(m):
        Hhat[i*m+pos,j] = hijbits[pos]

  H,c = reduce(Hhat)
  if not is_semi_systematic(H,c,mu,nu): return False

  assert c[:n-k-mu] == list(range(n-k-mu))
  if mu > 0: assert n-k-mu <= c[n-k-mu]
  assert all(c[i] <= i-mu+nu for i in range(n-k))

  H = copy(H)
  alphaprime = copy(alpha)
  for i in range(n-k-mu,n-k):
    if i != c[i]:
      print('swap %s %s' % (i,c[i]))
      Hi = copy(H.column(i))
      Hci = copy(H.column(c[i]))
      H.set_column(i,Hci)
      H.set_column(c[i],Hi)
      alphaprime[i],alphaprime[c[i]] = alphaprime[c[i]],alphaprime[i]

  assert H.matrix_from_columns(range(n-k)) == identity_matrix(n-k)
  T = H.matrix_from_columns(range(n-k,n))
  Gammaprime = g,alphaprime

  assert is_support(alphaprime)
  assert is_publickey(T)

  assert sorted(alphaprime) == sorted(alpha)
  assert alphaprime[:n-k-mu] == alpha[:n-k-mu]
  assert alphaprime[n-k-mu+nu:] == alpha[n-k-mu+nu:]
  
  c = tuple(c[n-k-mu:n-k])
  assert is_columnselection(c)
  result = (T,)+c+(Gammaprime,)

  if mu == 0: assert Gammaprime == Gamma
  if mu == 0: assert result == (T,Gamma)
  if mu == 0: assert result == SystematicMatGen(Gamma)
  return result


def Irreducible(d):
  assert is_bit_vector(d,sigma1*t)
  beta = [sum(d[sigma1*j+i]*zinFq^i for i in range(m)) for j in range(t)]
  assert all(parent(betaj) == Fq for betaj in beta)
  beta = sum(beta[j]*yinFqt^j for j in range(t))
  g = beta.minpoly()
  if g.degree() != t: return False
  assert is_allowed_polynomial(g)
  return g

def FieldOrdering(b):
  assert is_bit_vector(b,sigma2*q)
  a = [sum(ZZ(b[sigma2*j+i])*2^i for i in range(sigma2)) for j in range(q)]
  assert all(parent(ai) == ZZ for ai in a)
  assert all(ai >= 0 for ai in a)
  assert all(ai < 2^sigma2 for ai in a)

  if len(set(a)) < q: return False

  a = [(a[i],i) for i in range(q)]
  a.sort()

  for i in range(q-1):
    assert a[i][0] < a[i+1][0]

  alpha = []
  for i in range(q):
    alpha += [sum((1&(a[i][1]>>j))*zinFq^(m-1-j) for j in range(m))]

  assert is_fieldordering(alpha)
  return alpha

def SeededKeyGen(delta):
  while True:
    delta_str = "0x{0:064x}".format(int("".join([str(v) for v in delta]), 2))

    print(delta_str)

    f = open(args.fileseed, "w")
    f.write("\n".join(textwrap.wrap("{0:0256b}".format(int(delta_str,16)),32))+"\n")
    f.close()

    assert is_bit_vector(delta,l)
    E = vector_from_bytes(G(bytes_from_vector(delta)),n+sigma2*q+sigma1*t+l)
    deltaprime = E[n+sigma2*q+sigma1*t:]
    s = E[:n]


    d = E[n+sigma2*q:n+sigma2*q+sigma1*t]

    d = [d[i:i + sigma1] for i in range(0, len(d), sigma1)]
    
    # Need to reverse bit order!
    d_out = ["".join(str(i) for i in v)[::-1] for v in d]
    
    #print(d_out)
    
    d_out = "\n".join(d_out[::-1])
    
    f = open("sage_beta_rand.out", "w")
    f.write(d_out)
    f.write("\n")
    f.close()

    b = E[n:n+sigma2*q]

    b = [b[i:i + sigma2] for i in range(0, len(b), sigma2)]
    
    # Need to reverse bit order!
    b_out = "\n".join(["".join(str(i) for i in v)[::-1] for v in b])
    
    #print(b_out)
    
    f = open("sage_alpha_rand.out", "w")
    f.write(b_out)
    f.write("\n")
    f.close()

 
    alpha = FieldOrdering(E[n:n+sigma2*q])
    if alpha == False:
      delta = deltaprime
      sys.stderr.write("FiledOrdering fail; retry\n")
      continue

    g = Irreducible(E[n+sigma2*q:n+sigma2*q+sigma1*t])
    if g == False:
      delta = deltaprime
      sys.stderr.write("Irreducible fail; retry\n")
      continue

    Gamma = g,alpha[:n]
    MG = MatGen(Gamma)
    if MG == False:
      if args.fail:
        f = open("key_sage.out", "w")
        f.write("fail\n")
        f.close()
  
        return
      else:
        delta = deltaprime
        sys.stderr.write("MatGen fail; retry\n")
        continue


    T = MG[0]
    Gammaprime = MG[-1]
    c = MG[1:-1]
    assert MG == (T,)+c+(Gammaprime,)

    assert g == Gammaprime[0]
    alphaprime = Gammaprime[1]
    assert is_support(alphaprime)

    alpha = alphaprime+alpha[n:]

    assert is_bit_vector(delta,l)
    assert is_columnselection(c)
    assert is_allowed_polynomial(g)
    assert is_fieldordering(alpha)
    assert is_bit_vector(s,n)
    assert is_publickey(T)

    publickey = T
    privatekey = delta,c,g,alpha,s

    # fmt_m = "{0:0" + str(m) + "b}"
    # 
    # for c in reversed(g.coefficients()):
    #   print(fmt_m.format(c.integer_representation()), end='')
    # print()
    # 
    # for i, a in enumerate(alpha):
    #   print(i, fmt_m.format(a.integer_representation())[::-1])

    f = open(args.fileout, "w")
    f.write("\n".join(["".join([str(v) for v in r]) for r in T]) + "\n")
    f.close()

    return publickey,privatekey

if args.delta:
  delta = "{0:0256b}".format(int(args.delta))
  SeededKeyGen([GF(2)(int(v)) for v in delta])
else:
  SeededKeyGen([GF(2)(random.randint(0,1)) for i in range(l)])

