# This (sub-)makefile is part of the Classic McEliece hardware source code.
# 
# Copyright (C) 2022
#
# Authors: Norman Lahr <norman@lahr.email>
# 
# This source code is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# This source code is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this source code. If not, see <https://www.gnu.org/licenses/>. 

# Seeded Keygen Module

SEEDEDKEYGEN_SRC_PATH := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

# Include the parameter set definitions
include $(SEEDEDKEYGEN_SRC_PATH)/../../config/parameters.mk

# Include toolset
include $(SEEDEDKEYGEN_SRC_PATH)/../../config/tools.mk

# Name of the design's topmodule
TOPMODULE ?= seededkeygen

# List of the required submodules
SEEDEDKEYGEN_SUBMODULES := KEYGEN MEMORY_SINGLE SHAKE256 #REARRANGE_VECTOR

#
# Directory and path definitions
#
# Build directory
BUILD_DIR ?= $(SEEDEDKEYGEN_SRC_PATH)/build
# Directory pointing to the known answer tests
KAT_DIR ?= $(BUILD_DIR)/kat
# Directory for the performance results
RESULTS_DIR ?= $(BUILD_DIR)/results

# Source path for the common used modules
COMMON_SRC_PATH ?= $(SEEDEDKEYGEN_SRC_PATH)/../common
# Source path for the KAT generation script
KATGEN_SRC_PATH ?= $(SEEDEDKEYGEN_SRC_PATH)/../../kat

.SECONDEXPANSION:

#
# Parameter-specific build subdirectories
#
ifndef BUILD_DIR_TEMPLATE
define BUILD_DIR_TEMPLATE =

# Parameter-specific build directory root
BUILD_DIR_$(1) = $$(BUILD_DIR)/$(1)/$$(TOPMODULE)/$$(SYSTEMIZER)
# Parameter-specific source, testbench, and KAT directory
BUILD_DIR_SRC_$(1) = $$(BUILD_DIR_$(1))/src
BUILD_DIR_TB_$(1) = $$(BUILD_DIR_$(1))/testbench
BUILD_DIR_SYNTH_$(1) = $$(BUILD_DIR_$(1))/synthesis
BUILD_DIR_KAT_$(1) = $$(KAT_DIR/$(1))

endef

$(foreach par, $(PAR_SETS), $(eval $(call BUILD_DIR_TEMPLATE,$(par))))

endif

#
# Configurations and options of the module
#
# The available systemizers are organized in
# common/systemizer/modules.mk:
# SYSTEMIZER_SPEA_REDO
# SYSTEMIZER_SPGF2
# SYSTEMIZER_SPGF2_REDO
# SYSTEMIZER_SPGFX
# SYSTEMIZER_SPGFXA
SYSTEMIZER := SYSTEMIZER_SPEA_REDO

#
# Include makefiles for the submodules and KAT generation
#
include $(SEEDEDKEYGEN_SRC_PATH)/keygen/module.mk
include $(COMMON_SRC_PATH)/shake256/module.mk
include $(COMMON_SRC_PATH)/memory/module.mk
include $(KATGEN_SRC_PATH)/Makefile


#
# Sources definitions
#

# Add sources of the topmodule
SEEDEDKEYGEN_SRC = seeded_key_gen.v
SEEDEDKEYGEN_SRC += rearrange_vector.v

# Add the sources of the submodules
SEEDEDKEYGEN_SRC += $(foreach module,$(SEEDEDKEYGEN_SUBMODULES),$($(module)_SRC))

# Define comulated target for generated sources
.PHONY: sources
sources: $(addprefix sources-, $(PAR_SETS))

# Set of all sources
SEEDEDKEYGEN =


# Define the parameter-specific sources and targets
define SEEDEDKEYGEN_SRC_TEMPLATE =

# Define all required targets to be included in the modules above.
# The sort function filters double entries.
SEEDEDKEYGEN_$(1) = $(sort $(addprefix $$(BUILD_DIR_SRC_$(1))/,$(SEEDEDKEYGEN_SRC)))

# Collect all targets
SEEDEDKEYGEN += $$(SEEDEDKEYGEN_$(1))

# Define parameter-specific generated source targets
.PHONY: sources-$(1)
sources-$(1): $$(SEEDEDKEYGEN_$(1))

$$(BUILD_DIR_SRC_$(1)):
	mkdir -p $$@


#
# Sources generation
#
$$(BUILD_DIR_SRC_$(1))/seeded_key_gen.v: $$(SEEDEDKEYGEN_SRC_PATH)/seeded_key_gen.v
	cp $$< $$@

$$(BUILD_DIR_SRC_$(1))/rearrange_vector.v: $$(COMMON_SRC_PATH)/rearrange_vector.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call SEEDEDKEYGEN_SRC_TEMPLATE,$(par))))


ifeq (seededkeygen,${TOPMODULE})
#
# Testbench sources definitions
#

# Set the testbench source and data files of this module
SEEDEDKEYGEN_TB_SRC = seeded_key_gen_tb.v

# Define comulated target for generated sources
.PHONY: testbench
testbench: $(addprefix testbench-, $(PAR_SETS))

# Set of all testbench sources
SEEDEDKEYGEN_TB =


# Define the parameter-specific testbench targets
define SEEDEDKEYGEN_TB_TEMPLATE =

# Define all required targets to be included in the modules above
SEEDEDKEYGEN_TB_$(1) = $(addprefix $$(BUILD_DIR_TB_$(1))/,$(SEEDEDKEYGEN_TB_SRC))

# Collect all testbench targets
SEEDEDKEYGEN_TB += $$(SEEDEDKEYGEN_TB_$(1))

# Define parameter-specific testbench targets
.PHONY: testbench-$(1)
testbench-$(1): $$(SEEDEDKEYGEN_$(1)) $$(SEEDEDKEYGEN_TB_$(1))

# Make build directory for the testbench sources
$$(BUILD_DIR_TB_$(1)):
	mkdir -p $$@

#
# Sources generation
#
$$(BUILD_DIR_TB_$(1))/seeded_key_gen_tb.v: $$(SEEDEDKEYGEN_SRC_PATH)/testbench/seeded_key_gen_tb.v
	cp $$< $$@

endef

$(foreach par, $(PAR_SETS), $(eval $(call SEEDEDKEYGEN_TB_TEMPLATE,$(par))))


# Add the build directory as order-only-prerequisites to every source file target
$(SEEDEDKEYGEN) $(SEEDEDKEYGEN_TB): | $$(@D)


#
# Simulation targets and files
#

# Define Verilator-specific testbench source
SEEDEDKEYGEN_VERILATOR_TB_SRC = $(SEEDEDKEYGEN_SRC_PATH)/testbench/tb.cpp

# Parameter-specific simulator definitions
define SEEDEDKEYGEN_SIM_TEMPLATE =

# Required KAT files for encapsulation
SEEDEDKEYGEN_KAT_$(1) = \
	$$(KAT_DIR_$(1))/$(KAT_FILE_DELTA) \
	$$(KAT_DIR_$(1))/$(KAT_FILE_DELTA_PRIME) \
	$$(KAT_DIR_$(1))/$(KAT_FILE_PUBKEY)

# Set of all files to perform the simulation
SEEDEDKEYGEN_SIM_$(1) = \
	$$(SEEDEDKEYGEN_VERILATOR_TB_SRC) \
	$$(SEEDEDKEYGEN_TB_$(1)) \
	$$(SEEDEDKEYGEN_$(1)) \
	$$(SEEDEDKEYGEN_KAT_$(1))

# Simulation binaries
SEEDEDKEYGEN_VERILATOR_BIN_$(1) = $$(BUILD_DIR_TB_$(1))/seededkeygen_tb.bin

# Simulation performance files
SEEDEDKEYGEN_RESULT_TB_$(1) = $$(RESULTS_DIR)/cycles_profile_$(1)_$(TOPMODULE)_$(SYSTEMIZER).data

# Define parameter-specific simulation results
SIM_$(1) ?=
SIM_$(1) += $$(SEEDEDKEYGEN_RESULT_TB_$(1))

# Simulation binary target
$$(SEEDEDKEYGEN_VERILATOR_BIN_$(1)): $$(SEEDEDKEYGEN_SIM_$(1)) | $$(RESULTS_DIR)
	$(VERILATOR) \
	--unroll-count 1024 \
	-Wall -Wno-fatal \
	--trace \
	--Mdir $$(BUILD_DIR_TB_$(1))/obj_dir \
	--top-module seeded_key_gen_tb \
	-DM=$$(m_$(1)) \
	-DT=$$(t_$(1)) \
	-DK=$$(N_$(1)) \
	-DN_H=$$(N_H_elim_$(1)) \
	-DBLOCK_H=$$(BLOCK_H_elim_$(1)) \
	-DN_R=$$(N_R_elim) \
	-DBLOCK_R=$$(BLOCK_R_elim) \
	-DSIGMA1=$$(SIGMA1) \
	-DSIGMA2=$$(SIGMA2) \
	-DMEM_WIDTH=$$(mem_width_keygen_$(1)) \
	-DBANKS=$$(BANKS) \
	-DSRAM_DELAY=$$(SRAM_DELAY) \
	-DFILE_SEED=\"$$(KAT_DIR_$(1))/$$(KAT_FILE_DELTA_PRIME)\" \
	-DFILE_MEM_CONSTS=\"$$(BUILD_DIR_SRC_$(1))/mem_consts\" \
	-DFILE_PK_OUT=\"$$(BUILD_DIR_TB_$(1))/pubkey.out\" \
	-DFILE_VCD=\"$$(BUILD_DIR_TB_$(1))/wave.vcd\" \
	-DFILE_CYCLES_PROFILE=\"$$(SEEDEDKEYGEN_RESULT_TB_$(1))\" \
	-I$$(BUILD_DIR_SRC_$(1)) \
	-I$$(BUILD_DIR_TB_$(1)) \
	--cc \
	$$(BUILD_DIR_SRC_$(1))/clog2.v \
	$$(SEEDEDKEYGEN_TB_$(1)) \
	$$(filter-out $$(BUILD_DIR_SRC_$(1))/mem_consts $$(BUILD_DIR_SRC_$(1))/clog2.v,$$(SEEDEDKEYGEN_$(1))) \
	--exe $$< \
	--build \
	-j 8 \
	--threads 1 \
	--x-initial 0 \
	-o $$@

# Perform the simulation, record the performance data, and check the outcome.
$$(SEEDEDKEYGEN_RESULT_TB_$(1)): $$(SEEDEDKEYGEN_VERILATOR_BIN_$(1))
	$$<
	@diff $$(KAT_DIR_$(1))/$(KAT_FILE_PUBKEY) $$(BUILD_DIR_TB_$(1))/pubkey.out \
	&& echo "Test Passed!"

.PHONY: sim-$(1)
sim-$(1): $$(SIM_$(1))

endef

$(foreach par, $(PAR_SETS), $(eval $(call SEEDEDKEYGEN_SIM_TEMPLATE,$(par))))


# Define comulated target for generated sources
.PHONY: sim
sim: $(addprefix sim-, $(PAR_SETS))


#
# Synthesis targets and files
#
# Include vendor- and devices-specific definitions and recipes
include $(SEEDEDKEYGEN_SRC_PATH)/../../config/synthesis.mk

# Define the synthesis recipes for Xilinx devices
define SEEDEDKEYGEN_SYNTH_TEMPLATE =

# Define required report files
SEEDEDKEYGEN_RESULT_SYNTH_UTILIZATION_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_utilization_$(1)_$(2).txt
SEEDEDKEYGEN_RESULT_SYNTH_UTILIZATION_HIERARCHICAL_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_utilization_hierarchical_$(1)_$(2).txt
SEEDEDKEYGEN_RESULT_SYNTH_TIMING_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_timing_$(1)_$(2).txt
SEEDEDKEYGEN_RESULT_SYNTH_TIMING_SUMMARY_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_timing_summary_$(1)_$(2).txt
SEEDEDKEYGEN_RESULT_SYNTH_CLOCKS_$(1)_$(2) = $$(RESULTS_DIR)/$(TOPMODULE)_clocks_$(1)_$(2).txt

# Perform the synthesis
# #TODO (MAKE versions < 3.8 do not support mutli-target rules)
$$(SEEDEDKEYGEN_RESULT_SYNTH_CLOCKS_$(1)_$(2)): $$(SEEDEDKEYGEN_$(2)) $$(XILINX_RESOURCE_ANALYSIS_TCL) $$(XILINX_CONSTRAINTS_TIMING_$(1)) | $(RESULTS_DIR)
	$$(VIVADO) \
	-nojournal \
	-log $$(RESULTS_DIR)/synthesis_$(1)_$(2).log \
	-mode batch \
	-source $$(XILINX_RESOURCE_ANALYSIS_TCL) \
	-tclargs \
		"$$(BUILD_DIR_SRC_$(2))/clog2.v $(filter-out $$(BUILD_DIR_SRC_$(2))/clog2.v,$$(SEEDEDKEYGEN_$(2)))" \
		$$(XILINX_CONSTRAINTS_TIMING_$(1)) \
		seeded_key_gen \
		$$(XILINX_FPGA_MODEL_ID_$(1)) \
		$$(SEEDEDKEYGEN_RESULT_SYNTH_UTILIZATION_$(1)_$(2)) \
		$$(SEEDEDKEYGEN_RESULT_SYNTH_UTILIZATION_HIERARCHICAL_$(1)_$(2)) \
		$$(SEEDEDKEYGEN_RESULT_SYNTH_TIMING_$(1)_$(2)) \
		$$(SEEDEDKEYGEN_RESULT_SYNTH_TIMING_SUMMARY_$(1)_$(2)) \
		$$(SEEDEDKEYGEN_RESULT_SYNTH_CLOCKS_$(1)_$(2)) \
		"parameter_set=$$(id_$(2))" \
		FILE_MEM_CONSTS=$$(BUILD_DIR_SRC_$(2))/mem_consts

.PHONY: synthesis-$(1)-$(2)

synthesis-$(1)-$(2): $$(SEEDEDKEYGEN_RESULT_SYNTH_CLOCKS_$(1)_$(2))

# Define parameter-specific synthesis results
SYNTH_$(2) ?=
SYNTH_$(2) += $$(SEEDEDKEYGEN_RESULT_SYNTH_CLOCKS_$(1)_$(2))

endef

$(foreach model, $(XILINX_MODELS), \
	$(foreach par, $(PAR_SETS), \
		$(eval $(call SEEDEDKEYGEN_SYNTH_TEMPLATE,$(model),$(par))) \
	) \
)

# Define synthesis and results targets
define SEEDEDKEYGEN_RESULTS_TEMPLATE =

.PHONY: synthesis-$(1)
synthesis-$(1): $$(SYNTH_$(1))

.PHONY: results-$(1)
results-$(1): $$(SIM_$(1)) $$(SYNTH_$(1))

endef

$(foreach par, $(PAR_SETS), $(eval $(call SEEDEDKEYGEN_RESULTS_TEMPLATE,$(par))))


# Define comulated target for the synthesis
.PHONY: synthesis
synthesis: $(addprefix synthesis-, $(PAR_SETS))

# Define comulated target for all performance results
.PHONY: results
results: $(addprefix results-, $(PAR_SETS))

# Add the build directory as order-only-prerequisites to every source file target
$(SEEDEDKEYGEN) $(SEEDEDKEYGEN_TB): | $$(@D)

$(RESULTS_DIR):
	mkdir -p $@

endif
