/*
 * 
 *
 * Copyright (C) 2021
 * Author: Sanjay Deshpande <sanjay.deshpande@yale.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

//`timescale 1ns / 1ps

module seeded_key_gen
#( 

    parameter parameter_set = 1,
    
    
    parameter N = (parameter_set == 1)? 3488:
                  (parameter_set == 2)? 4608:
                  (parameter_set == 3)? 6688:
                  (parameter_set == 4)? 6960: 
                                        8192,
                                         
    parameter M = (parameter_set == 1)? 12:13,    
    
    parameter T = (parameter_set == 1)? 64:
                  (parameter_set == 2)? 96:
                  (parameter_set == 3)? 128:
                  (parameter_set == 4)? 119: 
                                        128,    
   
   
   // common parameters
    parameter K = N-M*T,
    parameter L = M*T,
    parameter Q = 2**M,
    parameter SIGMA1 = 16,
    parameter SIGMA2 = 32,
`ifdef SHARED
    parameter N_H = 16, // N in H elimination
    parameter BLOCK_H = 16, // BLOCK in H elimination
`else
    parameter N_H = T, // N in H elimination
    parameter BLOCK_H = 32, // BLOCK in H elimination
`endif
    parameter N_R = 2, // N in R elimination
    parameter BLOCK_R = 1, // BLOCK in R elimination

    
    parameter MEM_WIDTH = 32,
   
    parameter K_H_elim = N_H*((N+N_H-1)/N_H),
    parameter K_R_elim = N_R*((T+1+N_R-1)/N_R),
    
    parameter S_RAM_DEPTH = (N+(32-N%32)%32)/32,

`ifdef SHARED
   parameter FFT_mem_width = 1 << (M-`CLOG2(T+1)-1),
   parameter DATA = "",
`endif
    parameter SEED_SIZE = 256,
    parameter SEED_RAM_DEPTH = (SEED_SIZE +(32-SEED_SIZE%32)%32)/32,
    parameter fblock = (N == 3488 && T == 64  && M == 12)? 32'h400212A0:   
                       (N == 4608 && T == 96  && M == 13)? 32'h40041900:
                       (N == 6688 && T == 128 && M == 13)? 32'h40042320:
                       (N == 6960 && T == 119 && M == 13)? 32'h400423A0:
                       (N == 8192 && T == 128 && M == 13)? 32'h40042900:
                                                           32'hFFFFFFFF //invalid mode
)
(
    input clk,
    input rst,
    input delta_valid,
    input [31:0] delta,
    output reg done,
`ifndef SHARED
    output valid_vector,
`endif
    //shake interface ports
    output reg          shake_din_valid_kg,
    input wire          shake_din_ready_kg,
    output wire [31:0]  shake_din_kg,
    output wire         shake_dout_ready_kg,
    input wire [31:0]   shake_dout_scrambled_kg,
    output reg          shake_force_done_kg,
    input wire          shake_dout_valid_kg,
 
     // general inputs
    //input wire                                start, // general start signal
    //input wire [1:0]                          request, 
    // Simple instructions
    // 00: given two seeds, return a key pair
    
    // polynomial generation
    
    // S   
    input wire                               s_rd_en,
    input wire [`CLOG2(S_RAM_DEPTH)-1:0]     s_rd_addr, 
    output wire [32-1:0]                     s_rd_out,
    
    //
    input wire                                host_sk_poly_rd_en,
    input wire [`CLOG2(T*K_R_elim/N_R)-1:0]   host_sk_poly_rd_addr,
    output wire [M*N_R-1:0]                   sk_poly_dout,
    output wire                               sk_poly_gen_done,
    
    // permutation generation
    input wire                                host_perm_rd_en,
    input wire [M-1:0]                        host_perm_rd_addr,
    output wire [M-1:0]                       perm_dout,
    output wire                               perm_done,

`ifdef SHARED
    // shared fieldordering generation  (secret permutation generator)
    output wire                                 host_alpha_rand_wr_en_kg, 
    output wire [M-1:0]                         host_alpha_rand_wr_addr_kg,
    output wire [SIGMA2-1:0]                    host_alpha_rand_data_in_kg,
    output wire                                 sk_gen_start_kg,
    input  wire                                 perm_done_kg,
    input  wire                                 field_ordering_fail_kg,
    output wire                                 perm_rd_en_valid_kg,
    output wire [M-1:0]                         perm_rd_addr_valid_kg,
    input  wire [M-1:0]                         perm_dout_kg,
    input  wire [SIGMA2-1:0]                    perm_rand_out_kg,
  
    //ports for SHARED ADDITIVE FFT module 
    output wire fft_eva_start_kg,
    output wire [M*(T+1)-1:0] fft_coeff_in_kg,
    output wire fft_rd_en_kg,
    output wire [`CLOG2(T+1)-1:0] fft_rd_addr_kg,
    input wire [2*M*FFT_mem_width-1:0] fft_eva_dout_kg,
    input wire fft_eva_done_kg,
    
    input wire                                  host_pk_wr_en,
    input wire [`CLOG2(L*K_H_elim/N_H)-1:0]     host_pk_wr_addr, 
    input wire [N_H-1:0]                        host_pk_din,
`endif

    // PK generation
    input wire                                host_pk_rd_en,
    input wire [`CLOG2(L*K_H_elim/N_H)-1:0]   host_pk_rd_addr, 
    output wire [N_H-1:0]                     pk_dout,
    output reg                                pk_gen_done

    //testing purpose                              
//    input wire                               pk_gen_fail, 
//    input wire                               pk_gen_done                             
 
    
    );


reg [2:0] state = 0;
// Below are states
parameter s_initialize             =   0;
parameter s_wait_start             =   1;
parameter s_fill_s_mem             =   2;
parameter s_fill_alpha             =   3;
parameter s_fill_irr_poly_1        =   4;
parameter s_fill_irr_poly_2        =   5;
parameter s_start_kg               =   6;
parameter s_wait_done_or_fail      =   7;



// shake signals
reg [2:0] shake_input_type;
wire [31:0] din_shake;

// signals for the seed loading  
wire [31:0]seed_in;
reg [3:0]seed_addr;
reg seed_wr_en;
wire [31:0]seed_q;
reg initial_loading;
wire [3:0]addr_for_seed;
reg shake_dout_ready_kg_nab, shake_dout_ready_kg_deltap;

//s ports

wire [`CLOG2(S_RAM_DEPTH):0] s_addr;
reg  [`CLOG2(S_RAM_DEPTH):0] s_wr_addr;
wire [31:0] s_data_in;
reg s_wr_en;

// alpha generation signals
parameter ALPHA_BITS = SIGMA2*Q;
parameter RAND_ALPHA_MAX = (ALPHA_BITS + (32-ALPHA_BITS%32)%32)/32;
reg host_alpha_rand_wr_en;  
reg [M:0] host_alpha_rand_wr_addr;
wire [SIGMA2-1:0] host_alpha_rand_data_in;

// irreducible poly generation signals
parameter IRR_POLY_BITS = SIGMA1*T;
parameter IRR_POLY_MAX = (IRR_POLY_BITS + (16-IRR_POLY_BITS%16)%16)/16;
reg host_beta_rand_wr_en;  
reg [`CLOG2(T):0]  host_beta_rand_wr_addr;
reg [`CLOG2(T):0]  beta_counter;
wire [SIGMA1-1:0]  host_beta_rand_data_in; 
wire [SIGMA1-1:0]msb_shake_out;
reg [SIGMA1-1:0]lsb_shake_out;
reg poly_lsb;
reg shake_dout_valid_kg_reg;
reg beta_load_done;

//KEYGEN signal
reg [31:0] seed_q_reg;  
wire [31:0] shake_dout_rearranged;  
reg [31:0] shake_dout_rearranged_reg;  
wire [31:0] shake_dout_adjusted;  
wire done_kg;
reg start_kg;

wire [31:0]delta_bytes; 
wire pk_gen_fail;
reg [1:0] request;


assign shake_din_kg = din_shake;
assign  seed_in = (initial_loading)? delta: shake_dout_rearranged; 
assign addr_for_seed = (seed_wr_en)? l_seed_addr[3:0]: seed_addr[3:0]; 

genvar j;
generate
            for (j=0; j<8; j =j+1) begin 
                assign delta_bytes[j] = seed_in[7-j]; 
                assign delta_bytes[8+j] = seed_in[15-j]; 
                assign delta_bytes[16+j] = seed_in[23-j]; 
                assign delta_bytes[24+j] = seed_in[31-j]; 
            end    
endgenerate

//rearrange_vector delta_rearrange(.vector_in(seed_in),.vector_out(delta_bytes));

 mem_single #(.WIDTH(32), .DEPTH(SEED_RAM_DEPTH) ) mem_single_seed
 (
        .clock(clk),
        .data({delta_bytes[7:0],delta_bytes[15:8],delta_bytes[23:16],delta_bytes[31:24]}),
        .address(addr_for_seed),
        .wr_en(seed_wr_en),
        .q(seed_q)
 );
 
 
 assign shake_dout_ready_kg = shake_dout_ready_kg_nab |  shake_dout_ready_kg_deltap; //nab = N value, alpha, beta
                                                                                                      
 assign din_shake = (shake_input_type == 2'b001)? fblock: //including the generation of next seed  
                    (shake_input_type == 2'b010)? 32'h80000108: //seed length
                    (shake_input_type == 2'b011)? {seed_q[23:0],8'h40}: // appending 0x40 in the MSB of the delta value
                    (shake_input_type == 2'b000)? {seed_q[23:0], seed_q_reg[31:24]}:
                                                  {24'h000000, seed_q_reg[31:24]}; 
 
 always@(posedge clk)
 begin
    seed_q_reg <= seed_q;
 end 
 
 //s memory
 assign s_addr = s_wr_addr;
 assign s_data_in = ((N%32 != 0) && (s_addr == S_RAM_DEPTH-1))? {shake_dout_rearranged[31:N%32], 16'h0000}: shake_dout_rearranged;
 
 assign s_rd_out = s_out;
   mem_single #(.WIDTH(32), .DEPTH(S_RAM_DEPTH)) mem_s
 (
        .clock(clk),
        .data(s_data_in),
        .address(s_rd_en? s_rd_addr: s_addr),
        .wr_en(s_wr_en),
        .q(s_out)
 );  
 
 rearrange_vector SHAKE_dout_rearrange(.vector_in(shake_dout_scrambled_kg),.vector_out(shake_dout_rearranged));
 
 always@(posedge clk)
 begin
    if (shake_dout_valid_kg) begin
        shake_dout_rearranged_reg <= shake_dout_rearranged;
    end
 end
 assign shake_dout_adjusted = (N%32 == 0)? shake_dout_rearranged: {shake_dout_rearranged_reg[15:0],shake_dout_rearranged[31:16]};
 
 //alpha rand in
 wire [SIGMA2-1:0] host_alpha_rand_data_in_rev;
 assign host_alpha_rand_data_in_rev = shake_dout_adjusted;

genvar i;
generate
            for (i=0; i<SIGMA2; i =i+1) begin 
                assign host_alpha_rand_data_in[i] = host_alpha_rand_data_in_rev[SIGMA2-i-1];
            end    
endgenerate

// Irreducible poly rand in
 wire [SIGMA1-1:0] host_beta_rand_data_in_rev;
 assign host_beta_rand_data_in_rev = (poly_lsb)? lsb_shake_out: msb_shake_out;
 genvar k;
generate
            for (k=0; k<SIGMA1; k =k+1) begin 
                assign host_beta_rand_data_in[k] = host_beta_rand_data_in_rev[SIGMA1-k-1];
            end    
endgenerate
 
 assign msb_shake_out = shake_dout_adjusted[31:16];
 always@(posedge clk)
 begin
  if (shake_dout_valid_kg) begin
    lsb_shake_out <= shake_dout_adjusted[15:0];
  end
 end

    
// random bit distribution control logic    
always@(posedge clk)
begin
    if (rst) begin
        state <= s_initialize;
        shake_force_done_kg <= 1'b0;
        seed_is_loaded_in_shake_off <= 1'b0;
        s_wr_addr <= 0;
        host_alpha_rand_wr_addr <= 0;
        host_beta_rand_wr_addr <= IRR_POLY_MAX - 1;
        beta_counter <= 0;
        beta_load_done <= 1'b0;
        done <= 1'b0;
        pk_gen_done <= 1'b0;
    end
    
    else begin     
        if (state == s_initialize) begin
            done <= 1'b0;
            s_wr_addr <= 0;
            host_alpha_rand_wr_addr <= 0;
            host_beta_rand_wr_addr <= IRR_POLY_MAX - 1;
            shake_force_done_kg <= 1'b0;
            beta_load_done <= 1'b0;
            pk_gen_done <= 1'b0;
            beta_counter <= 0;
            if (seed_is_loaded_in_shake) begin
                state <= s_wait_start;
                seed_is_loaded_in_shake_off <= 1'b1;
            end
        end       
        
        else if (state == s_wait_start) begin        
            if (shake_dout_valid_kg) begin
               state <= s_fill_s_mem;
               seed_is_loaded_in_shake_off <= 1'b1;
               s_wr_addr <= s_wr_addr+1;
            end
        end 
        
        else if (state == s_fill_s_mem) begin
            seed_is_loaded_in_shake_off <= 1'b0;
            if (s_wr_addr <= S_RAM_DEPTH - 1) begin
                if (shake_dout_valid_kg) begin
                    s_wr_addr <= s_wr_addr + 1;
                    state <= s_fill_s_mem;
                end
            end
            else begin
                state <= s_fill_alpha;
                s_wr_addr <= 0;
            end
       end
       
       else if (state == s_fill_alpha) begin
           seed_is_loaded_in_shake_off <= 1'b0;
           if (host_alpha_rand_wr_addr <= RAND_ALPHA_MAX - 1) begin
                if (shake_dout_valid_kg) begin
                    host_alpha_rand_wr_addr <= host_alpha_rand_wr_addr + 1;
                    state <= s_fill_alpha;
                end
            end
            else begin
                state <= s_fill_irr_poly_1;
                host_alpha_rand_wr_addr <= 0;
            end       
       end
       
       else if (state == s_fill_irr_poly_1) begin
           seed_is_loaded_in_shake_off <= 1'b0;
//           if (host_beta_rand_wr_addr <= IRR_POLY_MAX - 1) begin
           if (beta_counter <= IRR_POLY_MAX - 1) begin
                if (shake_dout_valid_kg) begin
//                    host_beta_rand_wr_addr <= host_beta_rand_wr_addr + 1;
                    host_beta_rand_wr_addr <= host_beta_rand_wr_addr - 1;
                    beta_counter <= beta_counter + 1;
                    state <= s_fill_irr_poly_2;
                end
            end
            else begin
                state <= s_start_kg;
                beta_load_done <= 1'b1;
                host_beta_rand_wr_addr <= 0;
            end       
       end 
       
       else if (state == s_fill_irr_poly_2) begin
           seed_is_loaded_in_shake_off <= 1'b0;
//           if (host_beta_rand_wr_addr <= IRR_POLY_MAX - 1) begin
           if (beta_counter <= IRR_POLY_MAX - 1) begin
                if (shake_dout_valid_kg_reg) begin
//                    host_beta_rand_wr_addr <= host_beta_rand_wr_addr + 1;
                    host_beta_rand_wr_addr <= host_beta_rand_wr_addr - 1;
                    beta_counter <= beta_counter + 1;
                    state <= s_fill_irr_poly_1;
                end
            end
            else begin
                state <= s_start_kg;
                beta_load_done <= 1'b1;
                host_beta_rand_wr_addr <= IRR_POLY_MAX - 1;
                beta_counter <= 0;
            end       
       end       
            
       else if (state == s_start_kg) begin
            state <= s_wait_done_or_fail;
            done <= 1'b1;
       end
       
       else if (state == s_wait_done_or_fail) begin
            beta_load_done <= 1'b0;
            done <= 1'b0;
            if (done_kg && (!pk_gen_fail)) begin
                shake_force_done_kg <= 1'b1;
                pk_gen_done <= 1'b1;
                state <= s_initialize;
            end
            else if (pk_gen_fail) begin
                shake_force_done_kg <= 1'b0;
                pk_gen_done <= 1'b0;
                state <= s_initialize;
            end 
            
//            if (done_kg) begin
//                state <= s_initialize;
//            end  
       end
                      
        
    end 
    shake_dout_valid_kg_reg <= shake_dout_valid_kg;
end 

always@(state or shake_dout_valid_kg or shake_dout_valid_kg_reg  or delta_valid or seed_addr or s_wr_addr or host_beta_rand_wr_addr, beta_counter) 
begin
    case (state)
     s_initialize: begin
                    s_wr_en <= 1'b0;
                    host_alpha_rand_wr_en <= 1'b0;
                    host_beta_rand_wr_en <= 1'b0;
                    poly_lsb <= 1'b0;
                     start_kg <= 1'b0;
                     request <= 2'b11;
                  end 
            
      
      s_wait_start: begin
                        shake_dout_ready_kg_nab <= 1'b1;
                        if (shake_dout_valid_kg) begin
                            s_wr_en <= 1'b1;
                        end
                       
                    end
      s_fill_s_mem: begin 
                        shake_dout_ready_kg_nab <= 1'b1;
                        if (shake_dout_valid_kg) begin
                            s_wr_en <= 1'b1;
                        end
                        else begin
                            s_wr_en <= 1'b0;
                        end
                        
                    end
      
      s_fill_alpha: begin 
                        shake_dout_ready_kg_nab <= 1'b1;          
                        if (shake_dout_valid_kg) begin
                            host_alpha_rand_wr_en <= 1'b1;
                        end
                        else begin
                            host_alpha_rand_wr_en <= 1'b0;
                        end
                    end
       
       s_fill_irr_poly_1: begin 
                                 
                        if (shake_dout_valid_kg && (beta_counter <= IRR_POLY_MAX - 1)) begin
                            host_beta_rand_wr_en <= 1'b1;
                            poly_lsb <= 1'b0;
//                            shake_dout_ready_kg_nab <= 1'b1; 
                        end
                        else begin
                            host_beta_rand_wr_en <= 1'b0;
                            poly_lsb <= 1'b0;
//                            shake_dout_ready_kg_nab <= 1'b0; 
                        end
                        
//                        if  (N%32 !=0 && ((host_beta_rand_wr_addr <= IRR_POLY_MAX - 2)))  begin
                        if  (N%32 !=0 && ((beta_counter <= IRR_POLY_MAX - 2)))  begin
                              shake_dout_ready_kg_nab <= 1'b1;
                        end
//                        else if (N%32 ==0 && ((host_beta_rand_wr_addr <= IRR_POLY_MAX - 1)))  begin
                        else if (N%32 ==0 && ((beta_counter <= IRR_POLY_MAX - 1)))  begin
                              shake_dout_ready_kg_nab <= 1'b1;
                        end
                        else begin
                              shake_dout_ready_kg_nab <= 1'b0;
                        end
                        
                    end
       
       s_fill_irr_poly_2: begin 

                        if (shake_dout_valid_kg_reg && (beta_counter <= IRR_POLY_MAX - 1)) begin                                 
                            host_beta_rand_wr_en <= 1'b1;
                            poly_lsb <= 1'b1;
//                            shake_dout_ready_kg_nab <= 1'b1; 
                        end
                        else begin
                            host_beta_rand_wr_en <= 1'b0;
                            poly_lsb <= 1'b0;
//                            shake_dout_ready_kg_nab <= 1'b0; 
                        end
                        
//                        if  (N%32 !=0 && ((host_beta_rand_wr_addr <= IRR_POLY_MAX - 2)))  begin
                        if  (N%32 !=0 && ((beta_counter <= IRR_POLY_MAX - 2)))  begin
                              shake_dout_ready_kg_nab <= 1'b1;
                        end
//                        else if (N%32 ==0 && ((host_beta_rand_wr_addr <= IRR_POLY_MAX - 1)))  begin
                        else if (N%32 ==0 && ((beta_counter <= IRR_POLY_MAX - 1)))  begin
                              shake_dout_ready_kg_nab <= 1'b1;
                        end
                        else begin
                              shake_dout_ready_kg_nab <= 1'b0;
                        end
                        
                    end
       
       s_start_kg: begin
                        start_kg <= 1'b1;
                        request <= 2'b00;
                        s_wr_en <= 1'b0;
                        shake_dout_ready_kg_nab <= 1'b0;
                        host_beta_rand_wr_en <= 1'b0;
                   end
                    
       s_wait_done_or_fail : begin
                    start_kg <= 1'b0;
                    s_wr_en <= 1'b0;
                    shake_dout_ready_kg_nab <= 1'b0;
                    host_beta_rand_wr_en <= 1'b0;
                end
      default: begin
                    start_kg <= 1'b0;
                    s_wr_en <= 1'b0;
                    host_alpha_rand_wr_en <= 1'b0;
                    host_beta_rand_wr_en <= 1'b0;
                    request <= 2'b00; 
               end
    endcase

end 

//seed loading

reg [2:0] state_loading = 0;
parameter s_wait_new_load           = 0;
parameter s_loading_seed            = 1;
parameter s_wait_for_internal_seed  = 2;
parameter s_load_seed_shake         = 3;
parameter s_stall_load              = 4;
parameter s_wait_kg                 = 5;

reg seed_is_ready;
reg [3:0]l_seed_addr;

always@(posedge clk)
begin
    if (rst) begin
        state_loading <= s_wait_new_load;
        l_seed_addr <= 0;
        seed_is_ready <= 1'b0;
        shake_dout_ready_kg_deltap <= 1'b0;
    end
    
    else begin
        if (state_loading == s_wait_new_load) begin
            shake_dout_ready_kg_deltap <= 1'b0;
            if (delta_valid) begin   
                state_loading <= s_loading_seed;   
                l_seed_addr <= l_seed_addr + 1;
            end
        end
        
        else if (state_loading == s_loading_seed) begin
             if (done_kg && (!pk_gen_fail)) begin
                state_loading <= s_wait_new_load;
                seed_is_ready <= 1'b0;
            end
            else begin 
                if (l_seed_addr == SEED_RAM_DEPTH) begin
                    l_seed_addr <= 0;
                    state_loading <= s_wait_for_internal_seed;
                    seed_is_ready <= 1'b1;
                end
                else begin
                    if (delta_valid) begin 
                        state_loading <= s_loading_seed; 
                        l_seed_addr <= l_seed_addr + 1;  
                    end
                end
            end
        end
        
        else if (state_loading == s_wait_for_internal_seed) begin
            seed_is_ready <= 1'b0;
            if (done_kg && (!pk_gen_fail)) begin
                state_loading <= s_wait_new_load;
                shake_dout_ready_kg_deltap <= 1'b0;
            end 
            else if (beta_load_done) begin
                state_loading <= s_load_seed_shake;
                shake_dout_ready_kg_deltap <= 1'b0;
            end      
        end

        else if (state_loading == s_load_seed_shake) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_loading <= s_wait_new_load;
                seed_is_ready <= 1'b0;
                shake_dout_ready_kg_deltap <= 1'b0;
            end
            else begin
                if (l_seed_addr == SEED_RAM_DEPTH) begin
                    state_loading <= s_wait_for_internal_seed; 
                    seed_is_ready <= 1'b1;
                    l_seed_addr <= 0;
                    shake_dout_ready_kg_deltap <= 1'b0;
                end
                else begin
                    state_loading <= s_stall_load;
                    shake_dout_ready_kg_deltap <= 1'b1;              
                end
            end
        end 
        
        else if (state_loading == s_stall_load) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_loading <= s_wait_new_load;
                seed_is_ready <= 1'b0;
                shake_dout_ready_kg_deltap <= 1'b0;
            end
            else begin
                if (l_seed_addr < SEED_RAM_DEPTH) begin
                    if (shake_dout_valid_kg) begin
                        state_loading <= s_load_seed_shake; 
                        l_seed_addr <= l_seed_addr+1;  
                        shake_dout_ready_kg_deltap <= 1'b1; 
                    end 
                end
            end    
        end
        
        else if (state_loading == s_wait_kg) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_loading <= s_wait_new_load;
            end
            if (pk_gen_fail) begin
                state_loading <= s_wait_for_internal_seed;
            end       
        end        
      end
    
end

always@(state_loading or shake_dout_valid_kg or delta_valid or seed_addr) 
begin
    case (state_loading)
     s_wait_new_load: begin 
                        
                        initial_loading <= 1'b1;
                        if  (delta_valid == 1'b1) begin
                            seed_wr_en <= 1'b1;
                        end
                        else begin
                            seed_wr_en <= 1'b0;
                        end
                      end
     
     s_loading_seed: begin
                        if  (delta_valid == 1'b1) begin
                            seed_wr_en <= 1'b1;
                        end
                        else begin
                            seed_wr_en <= 1'b0;
                        end
                      end
     
     s_wait_for_internal_seed: begin
                                    initial_loading <= 1'b0;
                               end
     
     s_stall_load: begin
                        if  (shake_dout_valid_kg == 1'b1) begin
                            seed_wr_en <= 1'b1;
                        end
                        else begin
                            seed_wr_en <= 1'b0;
                        end
                      end
                       
      s_wait_kg: begin
//                    seed_wr_en <= 1'b0;
                 end
                  
      default: seed_wr_en <= 1'b0;
    endcase

end 


reg [2:0] state_shake = 0;
parameter s_init_shake                  =   0;
parameter s_wait_for_shake_out_ready    =   1;
parameter s_shake_out_w                 =   2;
parameter s_shake_in_w                  =   3;
parameter s_load_new_seed               =   4;
parameter s_stall_0                     =   5;
parameter s_append_40                   =   6;
parameter s_last_delta                  =   7;

reg [1:0] count_steps;
reg  seed_is_loaded_in_shake;
reg  seed_is_loaded_in_shake_off;

//shake parallel processing loading
always@(posedge clk)
begin

 //================start feeding the SHAKE with seed============================
     if (rst) begin
        state_shake <= s_init_shake;
        seed_addr <= 0;
        seed_is_loaded_in_shake <= 1'b0;
    end
    else begin
       
        if (state_shake == s_init_shake) begin
            count_steps <= 0;
            seed_addr <= 0;
            seed_is_loaded_in_shake <= 1'b0;
            if (seed_is_ready) begin
                state_shake <= s_shake_out_w;
            end
        end 
        
        else if (state_shake == s_shake_out_w) begin
             if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                seed_is_loaded_in_shake <= 1'b0;
             end
             else begin
                if (count_steps == 1) begin
                    state_shake <= s_shake_in_w;
                    count_steps <= 0;
                end 
                else begin
                    state_shake <= s_shake_out_w;
                    count_steps <= count_steps + 1;
                end
             end
        end
        
        else if (state_shake == s_shake_in_w) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                seed_is_loaded_in_shake <= 1'b0;
            end
             else begin
                if (count_steps == 1) begin
                    state_shake <= s_append_40;
                    count_steps <= 0;
                end 
                else begin
                    state_shake <= s_shake_in_w;
                    count_steps <= count_steps + 1;
                end
             end
        end
       
       
        else if (state_shake == s_append_40) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                seed_is_loaded_in_shake <= 1'b0;
            end
            else begin
                state_shake <= s_stall_0;
                seed_addr <= seed_addr + 1;
            end
        end
        
        else if (state_shake == s_load_new_seed) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                seed_is_loaded_in_shake <= 1'b0;
            end
            else begin
                if (seed_addr == SEED_RAM_DEPTH) begin
//                    state_shake <= s_wait_for_shake_out_ready;
                    state_shake <= s_last_delta;
                    seed_addr <= seed_addr + 1;
                    seed_is_loaded_in_shake <= 1'b0;
                end 
                else begin
                    state_shake <= s_stall_0;
                    seed_addr <= seed_addr + 1;
                end
            end
        end
        
        else if (state_shake == s_stall_0) begin
              if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                seed_is_loaded_in_shake <= 1'b0;
              end
              else begin
              if (seed_addr == SEED_RAM_DEPTH) begin
                    state_shake <= s_last_delta;
                    seed_is_loaded_in_shake <= 1'b0;
              end 
              else
                state_shake <= s_load_new_seed;
              end
        end
        
        else if (state_shake == s_last_delta) begin
            if (done_kg && (!pk_gen_fail)) begin
                state_shake <= s_init_shake;
                
            end
            else begin
                state_shake <= s_wait_for_shake_out_ready;
                seed_addr <= 0;
                seed_is_loaded_in_shake <= 1'b1;
            end
        end
       
       else if (state_shake == s_wait_for_shake_out_ready) begin
            if (done_kg && (!pk_gen_fail)) begin
                seed_is_loaded_in_shake <= 1'b0;
                state_shake <= s_init_shake;
            end
            else if (shake_dout_valid_kg) begin          
                if (seed_is_loaded_in_shake_off) begin
                    seed_is_loaded_in_shake <= 1'b0;
                    state_shake <= s_init_shake;
                end
            end 
       end
 //================end feeding the SHAKE with seed============================
 
    end 
end 
 
 
always@(state_shake or count_steps or seed_addr or shake_dout_valid_kg) 
begin
    case (state_shake)
     s_init_shake: begin
                    shake_din_valid_kg <= 1'b0; 
                    shake_input_type <= 3'b000;
                    
                  end 
            
     s_shake_out_w: begin
                       if (count_steps == 0) begin
                           shake_input_type <= 3'b001;
                           shake_din_valid_kg <= 1'b1;
                       end 
                       else begin
                           shake_din_valid_kg <= 1'b0; 
                       end 
                    end
     
     
     s_shake_in_w: begin
                       if (count_steps == 0) begin
                           shake_input_type <= 3'b010;
                           shake_din_valid_kg <= 1'b1;
                       end 
                       else begin
                           shake_din_valid_kg <= 1'b0; 
                       end 
                    end
      
      s_append_40: begin
                            shake_input_type <= 3'b011;
                            if (seed_addr < SEED_RAM_DEPTH) begin
                                shake_din_valid_kg <= 1'b1;
                            end
                            else begin
                                shake_din_valid_kg <= 1'b0;
                            end
                       end
                       
      s_load_new_seed: begin
                            shake_input_type <= 3'b000;
                            if (seed_addr < SEED_RAM_DEPTH) begin
                                shake_din_valid_kg <= 1'b1;
                            end
                            else begin
                                shake_din_valid_kg <= 1'b0;
                            end
                       end
       
       s_last_delta: begin
                            shake_input_type <= 3'b111;
                            shake_din_valid_kg <= 1'b1;
                            
                       end
      
      s_wait_for_shake_out_ready: begin
                            shake_din_valid_kg <= 1'b0;
                       end 
                      
      s_stall_0: begin
                            shake_din_valid_kg <= 1'b0;
                       end 
      default: shake_din_valid_kg <= 1'b0;
      
    endcase

end  
    
//    assign pk_gen_done = done_kg;
    
  key_gen #(.M(M), .T(T), .K(N), .N_H(N_H), .BLOCK_H(BLOCK_H), .N_R(N_R), .BLOCK_R(BLOCK_R), .SIGMA1(SIGMA1), .SIGMA2(SIGMA2), .MEM_WIDTH(MEM_WIDTH)) KEYGEN_module (
    .clk(clk),
    .start(start_kg),
    .request(request),
    //
    .host_beta_rand_wr_en(host_beta_rand_wr_en),
    .host_beta_rand_wr_addr(host_beta_rand_wr_addr),
    .host_beta_rand_data_in(host_beta_rand_data_in),
    //
    .host_sk_poly_rd_en(host_sk_poly_rd_en),
    .host_sk_poly_rd_addr(host_sk_poly_rd_addr),
    .sk_poly_dout(sk_poly_dout),
    .sk_poly_gen_done(sk_poly_gen_done),
    //
`ifndef SHARED
    .host_alpha_rand_wr_en(host_alpha_rand_wr_en),
    .host_alpha_rand_wr_addr(host_alpha_rand_wr_addr),
    .host_alpha_rand_data_in(host_alpha_rand_data_in),
`endif
    //
    .host_perm_rd_en(host_perm_rd_en),
    .host_perm_rd_addr(host_perm_rd_addr),
`ifdef SHARED
    .perm_dout(perm_dout_kg),
    .perm_done(perm_done_kg),
`else
    .perm_dout(perm_dout),
    .perm_done(perm_done),
`endif
    .host_pk_rd_en(host_pk_rd_en),
    .host_pk_rd_addr(host_pk_rd_addr),
    .pk_dout(pk_dout),
    .pk_gen_done(done_kg),
`ifdef SHARED
    .sk_gen_start(sk_gen_start_kg),
    .field_ordering_fail(field_ordering_fail_kg),
    .perm_rd_en_valid(perm_rd_en_valid_kg),
    .perm_rd_addr_valid(perm_rd_addr_valid_kg),
    .perm_rand_out(perm_rand_out_kg),

    //shared fft signals
    .fft_eva_start_kg(fft_eva_start_kg),
    .fft_coeff_in_kg(fft_coeff_in_kg),
    .fft_rd_en_kg(fft_rd_en_kg),
    .fft_rd_addr_kg(fft_rd_addr_kg),
    .fft_eva_dout_kg(fft_eva_dout_kg),
    .fft_eva_done_kg(fft_eva_done_kg),

    .host_pk_wr_en(host_pk_wr_en),
	.host_pk_wr_addr(host_pk_wr_addr),
	.host_pk_din(host_pk_din),
`endif
	.pk_gen_fail(pk_gen_fail)
  );

`ifdef SHARED
   defparam KEYGEN_module.DATA = DATA;
`else
   defparam KEYGEN_module.DATA = "";
`endif

`ifdef SHARED
assign host_alpha_rand_wr_en_kg = host_alpha_rand_wr_en;
assign host_alpha_rand_wr_addr_kg = host_alpha_rand_wr_addr;
assign host_alpha_rand_data_in_kg = host_alpha_rand_data_in;
`endif
endmodule
