/*
 * Copyright (C) 2021
 * Author: Sanjay Deshpande <sanjay.deshpande@yale.edu>
 *         Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

//`timescale 1ns/1ps

module seeded_key_gen_tb
       #(
           parameter L = `M*`T,
           parameter K_H_elim = `N_H*((`K+`N_H-1)/`N_H),
           parameter K_R_elim = `N_R*((`T+1+`N_R-1)/`N_R),
           parameter CLK_PERIOD = 2,
           parameter MODULE_NAME = "Seeded KeyGen"
       )
       (
           input wire clk,
           output wire pk_gen_done
       );

// input
reg rst = 1'b0;
reg delta_valid = 1'b0;
wire [32-1:0] seed_from_ram;
reg  [32-1:0] delta;

// output

// ram controls
reg [4:0] addr = 0;


// shake interfaces
wire          shake_din_valid_kg;
wire          shake_din_ready_kg;
wire [31:0]   shake_din_kg;
wire          shake_dout_ready_kg;
wire [31:0]   shake_dout_scramble;
wire          shake_force_done_kg;
wire          shake_dout_valid_kg;

reg host_sk_poly_rd_en = 1'b0;
reg [`CLOG2(`T*K_R_elim/`N_R)-1:0]   host_sk_poly_rd_addr = 0;
wire [`M*`N_R-1:0] sk_poly_dout;
wire sk_poly_gen_done;

reg host_perm_rd_en = 1'b0;
reg [`M-1:0] host_perm_rd_addr = 0;
wire [`M-1:0] perm_dout;
wire perm_done;

reg host_pk_rd_en = 0;
reg [`CLOG2(L*K_H_elim/`N_H)-1:0] host_pk_rd_addr = L*(L/`N_H);
wire [`N_H-1:0] pk_dout;
reg pk_gen_fail = 0;

wire pk_mod_done;


keccak_top shake_instance
           (
               .rst(rst),
               .clk(clk),
               .din_valid(shake_din_valid_kg),
               .din_ready(shake_din_ready_kg),
               .din(shake_din_kg),
               .dout_valid(shake_dout_valid_kg),
               .dout_ready(shake_dout_ready_kg),
               .dout(shake_dout_scramble),
               .force_done(0)
           );

seeded_key_gen # (
                   .M(`M),
                   .T(`T),
                   .N(`K),
                   .SIGMA1(`SIGMA1),
                   .SIGMA2(`SIGMA2),
                   .N_H(`N_H),
                   .BLOCK_H(`BLOCK_H),
                   .N_R(`N_R),
                   .BLOCK_R(`BLOCK_R),
                   .MEM_WIDTH(`MEM_WIDTH)
               )
               DUT  (
                   .clk(clk),
                   .rst(rst),
                   .delta_valid(delta_valid),
                   .delta(delta),
                   .done(),

                   //shake interface
                   .shake_din_valid_kg(shake_din_valid_kg),
                   .shake_din_ready_kg(shake_din_ready_kg),
                   .shake_din_kg(shake_din_kg),
                   .shake_dout_ready_kg(shake_dout_ready_kg),
                   .shake_dout_scrambled_kg(shake_dout_scramble),
                   .shake_force_done_kg(shake_force_done_kg),
                   .shake_dout_valid_kg(shake_dout_valid_kg),

                   //
                   .host_sk_poly_rd_en(host_sk_poly_rd_en),
                   .host_sk_poly_rd_addr(host_sk_poly_rd_addr),
                   .sk_poly_dout(sk_poly_dout),
                   .sk_poly_gen_done(sk_poly_gen_done),

                   //
                   .host_perm_rd_en(host_perm_rd_en),
                   .host_perm_rd_addr(host_perm_rd_addr),
                   .perm_dout(perm_dout),
                   .perm_done(perm_done),
                   .host_pk_rd_en(host_pk_rd_en),
                   .host_pk_rd_addr(host_pk_rd_addr),
                   .pk_dout(pk_dout),
                   .pk_gen_done(pk_mod_done)
                   //    .pk_gen_fail(pk_gen_fail)
               );


// Dump waveform.
// Caution: the seed keygen requires lots of cycles which results in a
// vcd-file approx. greater than 50GB.
// initial
//  begin
//    $dumpfile(`FILE_VCD);
//    $dumpvars(0, seeded_key_gen_tb);
//  end

/* Clock cycle count profiling */
integer f_cycles_profile;
//   time    time_decap_start;
time    time_sk_gen_start;
time    time_pk_gen_start;
integer time_pk_gen_right_start = -1;
reg [20*8+17*8-1:0] prefix = 0;

// Counter to control the test procedure
reg [5:0] ctr = 0;

// Set control signals and data
always @(posedge clk)
begin
    ctr <= ctr < 40 ? ctr + 1 : ctr;

    // reset for one cycle
    rst <= (ctr == 5) ? 1'b1 : 1'b0;

    delta <= seed_from_ram;

    // Time delta_valid with output from memory:
    addr <= (ctr > 21) && (ctr < 29) ? addr + 1 : addr;
    delta_valid <= (ctr > 22) && (ctr < 31) ? 1'b1 : 1'b0;
end

initial
begin
    f_cycles_profile = $fopen(`FILE_CYCLES_PROFILE,"w");
    $sformat(prefix, "[%s] [mceliece%0d%0d]", MODULE_NAME, DUT.N, DUT.T);
end

// Measure, diplay, and write out the times
always @(posedge DUT.KEYGEN_module.sk_gen_start)
begin
    $display("%s started.", MODULE_NAME);
    time_sk_gen_start = $time;
end

always @(posedge DUT.KEYGEN_module.sk_poly_gen_done)
begin
    $display("%s Generation of g(x) finished. (%0d cycles)", prefix, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "sk_poly_gen %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();
end

always @(posedge DUT.KEYGEN_module.perm_done)
begin
    $display("%s Permutation finished. (%0d cycles)", prefix, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "permutation %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();
end

// always @(posedge clk)
//   if (DUT.KEYGEN_module.host_beta_rand_wr_en)
//     $display("%b", DUT.KEYGEN_module.host_beta_rand_data_in);

always @(posedge DUT.KEYGEN_module.pk_gen_start)
begin
    time_pk_gen_start = $time;
    $display("%s Generation of the secret key finished. (%0d cycles)", prefix, ($time-time_sk_gen_start)/CLK_PERIOD);
    $fwrite(f_cycles_profile, "sk_gen %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
    $fflush();
end

always @(posedge DUT.KEYGEN_module.pk_gen_inst.success or posedge DUT.KEYGEN_module.pk_gen_inst.fail)
begin
    if(ctr > 10)
    begin
        time_pk_gen_right_start = $time;
        $display("%s Checking left part finished. (%0d cycles)", prefix, ($time-time_pk_gen_start)/CLK_PERIOD);
        $fwrite(f_cycles_profile, "check_left %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
        $fflush();
    end
end

always @(posedge DUT.KEYGEN_module.pk_gen_inst.elim_done)
begin
    if ((time_pk_gen_right_start > 1) && DUT.KEYGEN_module.pk_gen_inst.success)
    begin
        $display("%s Computing right part finished. (%0d cycles)", prefix, ($time-time_pk_gen_start)/CLK_PERIOD);
        $fwrite(f_cycles_profile, "compute_right %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
        $fflush();
    end
end

reg done = 1'b0;
assign pk_gen_done = done;

always  @(posedge clk)
  begin
    done <= host_pk_rd_addr == L*K_H_elim/`N_H-1;
  end

integer f2, i, j, k;

reg read_data = 1'b0;
reg read_data_del = 1'b0;

always
   @(posedge pk_mod_done)
   begin
     $display("%s Generation of public key finished. (%0d cycles)", prefix, ($time-time_pk_gen_start)/CLK_PERIOD);
     $fwrite(f_cycles_profile, "pk_gen %0d\n", ($time-time_pk_gen_start)/CLK_PERIOD);
     $fflush();

     $display("%s Generation of key pair finished. (%0d cycles)", prefix, ($time-time_sk_gen_start)/CLK_PERIOD);
     $fwrite(f_cycles_profile, "total %0d\n", ($time-time_sk_gen_start)/CLK_PERIOD);
     $fflush();

     f2 = $fopen(`FILE_PK_OUT,"w");
   end

reg [`CLOG2(L*K_H_elim/`N_H)-1:0] host_pk_rd_addr_del;

// Read out key data via main interface at the end of the computaiton.
always 
   @(posedge clk)
   begin
      read_data <= pk_mod_done ? 1'b1 : done ? 1'b0 : read_data;
      host_pk_rd_en <= pk_mod_done ? 1'b1 : done ? 1'b0 : host_pk_rd_en;

      host_pk_rd_addr_del <= host_pk_rd_addr;
      read_data_del <= read_data;

      host_pk_rd_addr <= pk_mod_done ? L*(L/`N_H) :
                         read_data ?
                         ( host_pk_rd_addr + L < L*(K_H_elim/`N_H) ? host_pk_rd_addr + L :
                           L*(L/`N_H) + (host_pk_rd_addr % L) + 1)
                         : host_pk_rd_addr;

      if (host_pk_rd_addr == L*(K_H_elim/`N_H))
         $display("done\n");

      if (read_data_del == 1'b1)
      begin
        if (host_pk_rd_addr_del < L*(K_H_elim/`N_H) - L)
          begin
            for (j = 0; j < `N_H; j = j+1)
              $fwrite(f2, "%b", pk_dout[j]);
          end
        else
          begin
            for (j = 0; j < `N_H - (K_H_elim - `K); j = j+1)
              $fwrite(f2, "%b", pk_dout[j]);
            $fwrite(f2, "\n");
          end
      end

      if (done)
        $fclose(f2);
   end



 mem_single #(.WIDTH(32), .DEPTH(8), .FILE(`FILE_SEED) ) mem_init_seed
            (
                .clock(clk),
                .data(0),
                .address(addr),
                .wr_en(0),
                .q(seed_from_ram)
            );

endmodule
